/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "rgBezier.h"
#include <rgStandard.h>

/****************************************************************************
 * BEGIN - Internal function prototypes.
 ****************************************************************************/
void rgGenBezierDegree1( int count, rgVector *cPoints, const int cStep,
	rgVector *bPoints, const int bStep);

void rgGenBezierDegree2( int count, rgVector *cPoints, const int cStep,
	rgVector *bPoints, const int bStep);

void rgGenBezierDegree3( int count, rgVector *cPoints, const int cStep,
	rgVector *bPoints, const int bStep);
/****************************************************************************
 * END - Internal function prototypes.
 ****************************************************************************/

int rgGenBezier( int count, rgVector *cPoints, const int cStep,
	rgVector *&bPoints, const int bStep, int level, int degree,
	bool autoAlloc_bPoints)
{
	/********************************************************************
	 * Check given parameters.
	 *
	 *	If the number of points is less then 0 or a null pointer to the
	 *	control points is given, then no bezier points are generated
	 *	and an error value is returned. bPoints is undefined if an
	 *	error occurs.
	 *
	 *	The given level must >= 0 and degree must be 1, 2 or 3. 
	 ********************************************************************/
	if( (count < 0) || (!cPoints) || level < 0) return -1;
	if( (degree < 1) || (degree > 3)) return -2;

	/********************************************************************
	 * If the number of control points is 0, then bPoints is set to NULL
	 * and 0 is returned.
	 ********************************************************************/
	if( 0 == count) {
		if( autoAlloc_bPoints) bPoints = NULL;
		return -120;
	}

	/********************************************************************
	 * Handle the trivial case of a bezier level of 0.
	 ********************************************************************/
	if( 0 == level) {
		if( autoAlloc_bPoints) bPoints = new rgVector[ count];
		//memcpy( bPoints, cPoints, count * sizeof(rgVector));
		rgVector *des = bPoints;
		rgVector *src = cPoints;
		for( int i = 0; i < count; i++) {
			*des = *src;
			des += bStep;
			src += cStep;
		}
		
		return count;
	}

	/********************************************************************
	 * Determine the number of bezier points that will be generated.
	 *
	 * Function for determining number of Bezier points generated:
	 *	f(x) = (2^level) * count - ((2^level) - 1) .
	 ********************************************************************/
	const int genCount = rgBezierPointCount( count, level);
	if( genCount <= 0) return -3;

	/********************************************************************
	 * Determine if this will be an "open condition" case bezier curve. 
	 ********************************************************************/
	//bool isOpenCondition = false;
	//if( cPoints[0] == cPoints[ count - 1]) isOpenCondition = true;

	/********************************************************************
	 * Allocate memory for the Bezier curve to be generated if the
	 * caller requested. Otherwise, no memory is allocated since bPoints
	 * should already point to allocated memory.
	 ********************************************************************/
	if( autoAlloc_bPoints) bPoints = new rgVector[ genCount];
	if( !bPoints) return -4;

	/*
	 * A "double buffereing" method is used to generate the final Bezier
	 * curve. By using two buffers (both large enough to store the entire
	 * bezier curve) the function can calculate the Bezier without the
	 * overhead of allocating and deleting memory. Also, by creating one
	 * large buffer and then interleaving it (for the two pointers) it is
	 * possible to make the Bezier curve generation cache coherent.
	 *
	 * TODO: It has been determined the interleaving is not optimal. This
	 *	is because there are more generated points then control
	 *	points so extra memory is allocated if interleaving is used.
	 *	Using two seperate packed arrays is more efficient since the
	 *	cpu will most likely be able to put both vertex arrays in
	 *	cache anyway.
	 */

	/* Allocate the memory to store both vertex pointers.
	 * NOTE: The two vertex arrays will be interleaved in the single
	 *	allocation inorder to make the Bezier calculation cahce
	 *	coherent.
	 */
	rgVector *workRam = new rgVector[ genCount * 2];
	rgVector *primary = workRam;
	if( !primary) {
		/* Due to an error, any memory allocated by this function must
		 * be freeed before returning.
		 */
		if( autoAlloc_bPoints) {
		/* It is not neccasary to set bPoints to NULL since the caller
		 * will be informed by a negative return value. However, it is
		 * done anyway to remined the caller that Bezier generation
		 * failed.
		 */
			delete [] bPoints;
			bPoints = NULL;
		}
		return -5;
	}
	//rgVector *secondary = primary + 1;
	rgVector *secondary = primary + genCount;

	/* Copy the control points into the primary vertex pointer. */
	/* TODO: Optimze this vertex copy operation since it is basically
	 *	a case of copying memory as fast as possible.
	 */
	int i;
	rgVector *p = primary;
	for( i = 0; i < count; i++) {

		*p = *cPoints;
		p++;
		cPoints += cStep;
	}

	int curCount = count;
	switch( degree) {
	case 1:
		//rgGenBezierDegree1( count, cPoints, bStep, bPoints, cStep,
		//	level);
		while( level--) {
			rgGenBezierDegree1( curCount, primary, 1, secondary,
				1);

			/* The points in the secondary buffer are now the new
			 * control points for the next Bezier level. Therefore,
			 * swap the primary and the secondary arrays. The
			 * verticies in the primary array are not needed
			 * anymore, so the primary array can be used to store
			 * the result of the next Bezier level. This process
			 * is repeated until the requested Bezier level is
			 * reached.
			 */
			p = primary;
			primary = secondary;
			secondary = p;

			curCount += curCount - 1;
		}
		break;
	case 2:	/* See "case 1" for doc. since this case works the same way. */
		while( level--) {
			rgGenBezierDegree2( curCount, primary, 1, secondary,
				1);

			p = primary;
			primary = secondary;
			secondary = p;

			curCount += curCount - 1;
		}
		break;
	case 3:	/* See "case 1" for doc. since this case works the same way. */
		while( level--) {
			rgGenBezierDegree3( curCount, primary, 1, secondary,
				1);

			p = primary;
			primary = secondary;
			secondary = p;

			curCount += curCount - 1;
		}
		break;
	default:
		std::cerr << "rgGenBezier: Degree " << degree << " not supported."
			"\n";
		exit( 0);
		break;	/* Execution never reaches this line */
	}

	/* Copy the generated verticies into the final vertex array.
	 * NOTE: The final generated points will ALWAYS be in the primary
	 *	vertex array.
	 */
	/* TODO: Optimze this vertex copy operation since it is basically
	 *	a case of copying memory as fast as possible.
	 */
	p = bPoints;
	for( i = 0; i < genCount; i++) {

		*p = primary[i];
		p += bStep;
	}

	/* Free the buffers used to generate the Bezier curve.
	 * NOTE: The secondary array was part of the primary array, so when
	 *	primary array is deleted, the secondary array goes with it
	 *	(it is deleted also).
	 */
	delete [] workRam;

	return genCount;
}


void rgGenBezierDegree1( int count, rgVector *cPoints, const int cStep,
	rgVector *bPoints, const int bStep)
{

}

void rgGenBezierDegree2( int count, rgVector *cPoints, const int cStep,
	rgVector *bPoints, const int bStep)
{

}

void rgGenBezierDegree3( int count, rgVector *cPoints, const int cStep,
	rgVector *bPoints, const int bStep)
{
	/********************************************************************
	 * Cubic Bezier, degree = 3 
	 * Uniform subdivision, u = 0.5
	 ********************************************************************/
	int i;
	rgVector t;
	const int iStop = count - (count % 3);
	//int j = 0;

	/* The compiler "should" optimize these constant integers so that the
	 * pointer dereferencing is done quickly in the "for" loop.
	 */
	const int c1Step = 1 * cStep;
	const int c2Step = 2 * cStep;
	const int c3Step = 3 * cStep;
	
	const int b1Step = 1 * bStep;
	const int b2Step = 2 * bStep;
	const int b3Step = 3 * bStep;
	const int b4Step = 4 * bStep;
	const int b5Step = 5 * bStep;
	const int b6Step = 6 * bStep;

	rgVector *cptr = cPoints;
	rgVector *bptr = bPoints;
	for( i = 0; i < iStop; i += 3) {
		/*
		bPoints[j] = cPoints[i];
		bPoints[j + 1] = 0.5f * (cPoints[i] + cPoints[i+1]);
		bPoints[j + 5] = 0.5f * (cPoints[i + 2] + cPoints[i + 3]);

		t = 0.5f * (cPoints[i + 1] + cPoints[i + 2]);
		bPoints[j + 2] = 0.5f * (t + bPoints[j + 1]);
		bPoints[j + 4] = 0.5f * (t + bPoints[j + 5]);
		bPoints[j + 3] = 0.5f * (bPoints[j + 2] + bPoints[j + 4]);
		*/

		bptr[ 0] = cptr[0];
		bptr[ b1Step] = 0.5f * (cptr[0] + cptr[ c1Step]);
		bptr[ b5Step] = 0.5f * (cptr[ c2Step] + cptr[ c3Step]);

		t = 0.5f * (cptr[ c1Step] + cptr[ c2Step]);
		bptr[ b2Step] = 0.5f * (t + bptr[ b1Step]);
		bptr[ b4Step] = 0.5f * (t + bptr[ b5Step]);
		bptr[ b3Step] = 0.5f * (bptr[ b2Step] + bptr[ b4Step]);
	
		//j += 6;
		cptr += c3Step;	/* cptr += 3 * cStep ( move 3 elements ) */
		bptr += b6Step;	/* bptr += 6 * bStep ( move 6 elements ) */
	}
	bptr[ 0] = cptr[0];
	

	
	/********************************************************************
	 * Finish remaining points (if any)
	 ********************************************************************/
	switch( count - iStop) {
	case 3:
		//rgGenBezierDegree2( 3, &cPoints[i], &bPoints[j], level); 
		rgGenBezierDegree2( 3, cptr, cStep, bptr, bStep);
		break;
	case 2:
		//rgGenBezierDegree1( 2, &cPoints[i], &bPoints[j], level);
		rgGenBezierDegree1( 2, cptr, cStep, bptr, bStep);
		break;
	}
}


int rgBezierPointCount( const int pCount, const int level)
{
	/* TODO: Add check if genCount will be too large. */
	int genCount =
		((int)pow( 2, level) * pCount) - ((int)pow( 2, level) - 1);

	return genCount;
}


