/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgShadow.h"
#include <rg3DRom.h>
#include <rgOpenGL.h>

rgShadowVolume::rgShadowVolume()
{
	m_caster = 0;

	m_shadow_cast_length = 1000.0f;

	m_vert = 0;
	m_tri = 0;
	m_edge_table = 0;
	m_quad = 0;
	m_tri_cap = 0;

	m_vcount = 0;
	m_tcount = 0;
	m_esize = 0;
	m_qcount = 0;
	m_tri_cap_count = 0;
}

rgShadowVolume::~rgShadowVolume()
{
	clear();
}

void rgShadowVolume::clear()
{
	clear_vertex_list();
	clear_tri_list();
	clear_edge_table();
	clear_quad_list();
	clear_tri_cap_list();
	m_caster = 0;
}

void rgShadowVolume::create( const sk3DRomModel *mdl)
{
	if( !(m_caster = mdl)) return;
	if( create_vertex_list()) return;
	if( create_tri_list()) return;
	if( create_edge_table()) return;
	if( create_quad_list()) return;
	if( create_tri_cap_list()) return;
	update_vertices();
	update_triangle_normals();
}

int rgShadowVolume::create_vertex_list()
{
	clear_vertex_list();
	if( (m_vcount = int( m_caster->v.size())) < 3) {
		m_vcount = 0;
		return -1;
	}
	if( !(m_vert = new rgVector[ m_vcount])) return -2;
	return 0;
}

void rgShadowVolume::update_vertices()
{
	if( !ready_for_use()) return;
	rgVector *v = m_vert;
	for( int i = 0; i < m_vcount; i++)
		*(v++) = m_caster->get_tv_vertex( i);
}

int rgShadowVolume::create_tri_list()
{
	clear_tri_list();
	if( (m_tcount = int( m_caster->f.size())) <= 0) {
		m_tcount = 0;
		return -1;
	}
	if( !(m_tri = new rgShadowTri[m_tcount])) return -2;
	for( int i = 0; i < m_tcount; i++) {
		m_tri[i].vid[0] = m_caster->f[i].vid[0];
		m_tri[i].vid[1] = m_caster->f[i].vid[1];
		m_tri[i].vid[2] = m_caster->f[i].vid[2];
	}
	return 0;
}

void rgShadowVolume::update_triangle_normals()
{
	if( !ready_for_use()) return;
	for( int i = 0; i < m_tcount; i++) {
		// TODO: Perhaps the sk3DRomModel "mdl" has already calculated
		//	facet normals, thus it would be optimal to just copy
		//	in the normals from "mdl" rather than calculating them
		//	here.
		// XXX: The calculated normal does not have to be unit normal.
		//	The normal is used simply for testing the view
		//	direction of model/geometry triangles.
		m_tri[i].normal = rgCalcTriangleNormal(
			m_vert[ m_tri[i].vid[0]],
			m_vert[ m_tri[i].vid[1]],
			m_vert[ m_tri[i].vid[2]]);
	}
}

// XXX: The size of the "m_tri_cap" array is "m_tcount * 2". This is because
//	every triangle that is a capping triangle has a corresponding "back
//	capping" triangle. All triangles can be capping triangles, meaning that
//	there can be (m_tcount * 2) capping triangles (capping triangles plus
//	their back capping counter parts).
int rgShadowVolume::create_tri_cap_list()
{
	clear_tri_cap_list();
	if( m_tcount <= 0) return -1;
	if( !(m_tri_cap = new rgShadowTriCap[ m_tcount * 2])) return -2;
	return 0;
}

int rgShadowVolume::create_quad_list()
{
	clear_quad_list();
	if( m_tcount <= 0) return -1;
	if( !(m_quad = new rgShadowQuad[ m_tcount * 3])) return -2;
	return 0;
}

int rgShadowVolume::create_edge_table()
{
	clear_edge_table();
	m_esize = ((m_vcount * m_vcount) / 2) - (m_vcount / 2);
	if( !(m_edge_table = new rgShadowEdgeEntry[ m_esize])) return -1;
	return 0;
}

void rgShadowVolume::build_ortho_silhouette( const rgVector w)
{
	memset( m_edge_table, 0, m_esize * sizeof( rgShadowEdgeEntry));
	rgShadowTri *t = m_tri;
	for( int i = 0; i < m_tcount; i++) {
		// XXX: The test in the 'if" statment is "critical" to
		//	producing a proper shadow volume.
		if( w.dot( t->normal) < 0) {
			mark_triangle_edges_as_used( t, i);
		} else t->casts_shadow = 0;
		t++;
	}
	build_ortho_quad_list( w * m_shadow_cast_length);
}

void rgShadowVolume::build_projected_silhouette( const rgVector p)
{
	memset( m_edge_table, 0, m_esize * sizeof( rgShadowEdgeEntry));
	rgShadowTri *t = m_tri;
	for( int i = 0; i < m_tcount; i++) {
		// XXX: The test in the 'if" statment is "critical" to
		//	producing a proper shadow volume.
		if( (m_vert[ t->vid[0]] - p).dot( t->normal) < 0)
			mark_triangle_edges_as_used( t, i);
		else t->casts_shadow = 0;
		t++;
	}
	build_projected_quad_list( p);
}

void rgShadowVolume::mark_triangle_edges_as_used( rgShadowTri *t,
	const int tid)
{
	mark_edge_as_used( t->vid[0], t->vid[1],
		tid, RGSHADOW_VOLUME_V0V1_ORDER);
	mark_edge_as_used( t->vid[1], t->vid[2],
		tid, RGSHADOW_VOLUME_V1V2_ORDER);
	mark_edge_as_used( t->vid[2], t->vid[0],
		tid, RGSHADOW_VOLUME_V2V0_ORDER);
	t->casts_shadow = 1;
}

// XXX: Array element (a, b) == array element (b, a).
void rgShadowVolume::mark_edge_as_used( int a, int b, const int tid,
	const int order)
{
	if( a < b) rgSwap( a, b);
	const int i = rg_lower_triangle_array_access( a, b);
	const int tmp = ((m_vcount * m_vcount) / 2) - (m_vcount / 2);
	if( i >= tmp) {
		std::cerr << "mark_edge_as_used: "
			"a = " << a << ", b = " << b << 
			", i = " << i << ", size = " << tmp << "\n";
		std::cerr << "Raising SIGSEGV -->";
		*((int *)0) = 0;
	}
	if( m_edge_table[i].used <= 1) {
		m_edge_table[i].used++;
		m_edge_table[i].tid = tid;
		m_edge_table[i].order = order;
	}
}

void rgShadowVolume::build_projected_quad_list( const rgVector p)
{
	m_qcount = 0;
	rgShadowEdgeEntry *e = m_edge_table;
	for( int row = 1; row < m_vcount; row++) {
		for( int col = 0; col < row; col++) {
			if( 1 == e[col].used) {
				build_projected_quad( e + col,
					m_quad + m_qcount, p);
				m_qcount++;
			}
		}
		e += row;
	}
	build_projected_capping_triangles( p);
}

void rgShadowVolume::build_ortho_quad_list( const rgVector w)
{
	m_qcount = 0;
	rgShadowEdgeEntry *e = m_edge_table;
	for( int row = 1; row < m_vcount; row++) {
		for( int col = 0; col < row; col++) {
			if( 1 == e[col].used) {
				build_ortho_quad( e + col, m_quad + m_qcount,
					w);
				m_qcount++;
			}
		}
		e += row;
	}
	build_ortho_capping_triangles( w);
}

void rgShadowVolume::build_projected_capping_triangles( const rgVector p)
{
	rgShadowTri *t = m_tri;
	m_tri_cap_count = 0;
	for( int i = 0; i < m_tcount; i++) {
		if( t->casts_shadow) {
			rgShadowTriCap *cap = m_tri_cap + m_tri_cap_count;
			cap->v[0] = m_vert[ t->vid[0]];
			cap->v[1] = m_vert[ t->vid[1]];
			cap->v[2] = m_vert[ t->vid[2]];

			// XXX: Back capping triangles face in the opposite
			//	direction to their "front" capping counter
			//	parts.
			rgShadowTriCap *cap_back = cap + 1;
			cap_back->v[0] = cap->v[2] + (cap->v[2] - p)
				* m_shadow_cast_length;
			cap_back->v[1] = cap->v[1] + (cap->v[1] - p)
				* m_shadow_cast_length;
			cap_back->v[2] = cap->v[0] + (cap->v[0] - p)
				* m_shadow_cast_length;

			m_tri_cap_count += 2;
		}
		t++;
	}
}

void rgShadowVolume::build_ortho_capping_triangles( const rgVector w)
{
	rgShadowTri *t = m_tri;
	m_tri_cap_count = 0;
	for( int i = 0; i < m_tcount; i++) {
		if( t->casts_shadow) {
			rgShadowTriCap *cap = m_tri_cap + m_tri_cap_count;
			cap->v[0] = m_vert[ t->vid[0]];
			cap->v[1] = m_vert[ t->vid[1]];
			cap->v[2] = m_vert[ t->vid[2]];

			// XXX: Back capping triangles face in the opposite
			//	direction to their "front" capping counter
			//	parts.
			rgShadowTriCap *cap_back = cap + 1;
			cap_back->v[0] = cap->v[2] + w;
			cap_back->v[1] = cap->v[1] + w;
			cap_back->v[2] = cap->v[0] + w;

			m_tri_cap_count += 2;
		}
		t++;
	}
}

void rgShadowVolume::build_quad( const rgShadowEdgeEntry *e,
	rgShadowQuad *q) const
{
	switch( e->order) {
	case RGSHADOW_VOLUME_V0V1_ORDER:
		q->v[0] = m_vert[ m_tri[ e->tid].vid[0]];
		q->v[1] = m_vert[ m_tri[ e->tid].vid[1]];
		break;

	case RGSHADOW_VOLUME_V1V2_ORDER:
		q->v[0] = m_vert[ m_tri[ e->tid].vid[1]];
		q->v[1] = m_vert[ m_tri[ e->tid].vid[2]];
		break;

	case RGSHADOW_VOLUME_V2V0_ORDER:
		q->v[0] = m_vert[ m_tri[ e->tid].vid[2]];
		q->v[1] = m_vert[ m_tri[ e->tid].vid[0]];
		break;	
	}
}

void rgShadowVolume::build_ortho_quad( const rgShadowEdgeEntry *e,
	rgShadowQuad *q, const rgVector w)
{
	build_quad( e, q);
	q->v[2] = q->v[1] + w;
	q->v[3] = q->v[0] + w;
}

void rgShadowVolume::build_projected_quad( const rgShadowEdgeEntry *e,
	rgShadowQuad *q, const rgVector p)
{
	build_quad( e, q);
	q->v[2] = q->v[1] + (q->v[1] - p) * m_shadow_cast_length;
	q->v[3] = q->v[0] + (q->v[0] - p) * m_shadow_cast_length;
}

void rgShadowVolume::render() const
{
	if( !m_quad) return;
	if( !m_tri_cap) return;
	glBegin( GL_TRIANGLES);
	render_quads_using_triangles();
	render_capping_triangles();
	glEnd();
}

// XXX: The quads are rendered using two triangles. This is an optimization
//	so that a state change is not needed to render quads and triangle
//	caps. Otherwise quads and triangles would have to be rendered
//	separatley with an OpenGL state change.
void rgShadowVolume::render_quads_using_triangles() const
{
	for( int i = 0; i < m_qcount; i++) {
		// FIXME: Shadow quads are backwards. Rendering from index
		//	3 to 0, but should be 0 to 3.
		glVertex( m_quad[i].v[3]);
		glVertex( m_quad[i].v[2]);
		glVertex( m_quad[i].v[1]);
	
		glVertex( m_quad[i].v[1]);
		glVertex( m_quad[i].v[0]);
		glVertex( m_quad[i].v[3]);
	}
}

void rgShadowVolume::render_capping_triangles() const
{
	rgShadowTriCap *t = m_tri_cap;
	for( int i = 0; i < m_tri_cap_count; i++) {
		glVertex( t->v[0]);
		glVertex( t->v[1]);
		glVertex( t->v[2]);
		t++;
	}
}


