/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <iostream>
#include <vector>
#include <io.h>
#include <time.h>
#include <string.h>

#include <windows.h>


class MainData
{
public:
	bool show_banner;
	bool update_only_if_new;	// Always update by default
	bool verbose;

	std::vector <char *> clist;		// List of files to copy
	char *des_dir;

	MainData() {
		show_banner = true;
		update_only_if_new = false;
		verbose = false;

		des_dir = 0;	// Use the current directory by default
	}
};

void get_file_name_only( std::string &des, std::string &src);
void process_args( const int argc, char *argv[], MainData &m);
int file_date_compare( const char *file_a, const char *file_b);
bool file_is_out_of_date( const char *old, const char *cur);

#define MAX_BUF_SIZE	4096
char buf[MAX_BUF_SIZE];

int main( int argc, char *argv[])
{
	buf[ MAX_BUF_SIZE - 1] = '\0';

	// Process the user arguments.
	MainData m;
	process_args( argc, argv, m);
	
	// Print banner (unless disabled).
	if( m.show_banner) {
		std::cout << "rgSDK \"cp\" simulator, version = 1.0\n";
	}

	// Perform the copy.
	if( m.clist.empty()) {
		std::cout << "Nothing to do. Type \"" << argv[0]
			<< " -h\" for help\n";
		return 0;
	}

	bool do_copy;
	while( !(m.clist.empty())) {

		do_copy = true;

		if( m.des_dir) {
			// Output directory was given. Therefore concatinate
			// the output directory with the copy target file name
			// so that a file relative to the output directory can
			// be referenced.

			// First add the output directory portion.
			strncpy( buf, m.des_dir, MAX_BUF_SIZE - 1);

			// If no final slash was given for the output directory
			// then add one.
			if( '\\' != m.des_dir[ strlen( m.des_dir) - 1])
				strncat( buf, "\\", MAX_BUF_SIZE - 1);

			// Finally concat the file name portion.
			std::string temp, temp2;
			get_file_name_only( temp, (temp2 = m.clist[0]));
			strncat( buf, temp.c_str(), MAX_BUF_SIZE - 1);

		} else {
			strncpy( buf, m.clist[0], MAX_BUF_SIZE - 1);
		}

		if( m.update_only_if_new) {
			// If the file is not out of date, then do not copy it.
			if( !file_is_out_of_date( buf, m.clist[0]))
				do_copy = false;
		}

		// Copy the file now (but do not copy over exact same file).
		if( do_copy) {
			if( !CopyFile( m.clist[0], buf, FALSE)) {
				std::cerr << "Fatal Error: Failed while "
					"copying file \""
					<< m.clist[0] << "\" to destination \""
					<< buf
					<< "\"\n";
				return -2;
			}

			if( m.verbose) std::cout << m.clist[0] << " copied\n";
		} else {
			if( m.verbose) std::cout << m.clist[0] << " skipped\n";
		}


		m.clist.erase( m.clist.begin());
	}

	return 0;
}

void get_file_name_only( std::string &des, std::string &src)
{
	des = "";
	const char *s = src.c_str();
	const int len = src.length();
	for( int i = len - 1; i >= 0; i--) {
		switch( s[i]) {
		case '\\':
		case '/':
			for( i++; i < len; i++) des += s[i];
			return;
		}
	}
}

bool file_is_out_of_date( const char *old, const char *cur)
{
	switch( file_date_compare( old, cur)) {
	case -2:	// "old" does not exist, so is out of date
	case 2:		// "old" is older, so is out of date
		return true;

	case -3:	// "cur" does not exist, this is an error
	case -1:	// Neither file exists, this is an error
		std::cerr << "Fatal Error: File \"" << cur << "\" does not "
			"exist.\n";
		exit( -3);

	case 1:		// File "old" is newer, so is not out of date
	case 0:		// Both files have equal time, so "old" is current
	default:	// Other results are interpreted as "not out of date"
		return false;
	}
}

int file_date_compare( const char *file_a, const char *file_b)
{
	// Open the first file and get information about it.
	_finddata_t fd_a;
	const long result_a = _findfirst( file_a, &fd_a);

	// Open the second file and get information about it.
	_finddata_t fd_b;
	const long result_b = _findfirst( file_b, &fd_b);
	
	//std::cerr << "file_a = " << file_a << ", file_b = " << file_b << "\n";
	//std::cerr << "resA = " << result_a << ", resB = " << result_b << "\n";

	// If neither file exists, return -1.
	if( (-1 == result_a) && (-1 == result_b)) return -1;

	// If file "a" does not exist, but file "b" does exist, then return -2.
	if( (-1 == result_a) && (-1 != result_b)) return -2;

	// If file "b" does not exist, but file "a" does exist, then return -3.
	if( (-1 != result_a) && (-1 == result_b)) return -3;

	// Compare the two existing files.
	const double t = difftime( fd_a.time_write, fd_b.time_write);
	
	if( 0.0 == t) return 0;		// Files have equal modification times
	else if( t > 0.0) return 1;	// File "a" is newer
	return 2;			// File "b" is newer
}

void process_args( const int argc, char *argv[], MainData &m)
{
	// NOTE: Skip argument 0 (the name of the executed program).
	for( int i = 1; i < argc; i++) {

		// If the argument is a command
		if( '-' == (argv[i])[0]) {
			if( strcmp( &((argv[i])[1]), "no_banner") == 0) {
				m.show_banner = false;

			} else if( (stricmp( &((argv[i])[1]), "h") == 0)
				|| (stricmp( &((argv[i])[1]), "help") == 0)
				|| (strcmp( &((argv[i])[1]), "?") == 0)) {
				
				std::cout <<
					"Options:\n"
					"-u\t\t- Only copy (update) out of date files\n"
					"-verbose\t- List files as they are copied or skipped\n"
					"-no_banner\t- Do not display program banner\n"
					"-h\t\t- Display this help screen\n"
					"\n";
				exit( 0);

			} else if( strcmp( &((argv[i])[1]), "u") == 0) {
				m.update_only_if_new = true;

			} else if( strcmp( &((argv[i])[1]), "verbose") == 0) {
				m.verbose = true;

			} else {
				std::cerr << "Fatal Error: Unknown argument \""
					<< argv[i] << "\"\n";
				exit( -1);
			}

		// Else, the argument is a file to copy or (if it is the last
		// argument) the destination directory.
		} else {

			if( i == (argc - 1)) {
				// The current argument is the destination directory.
				m.des_dir = argv[i];

			} else {
				// The current argument is a file to copy.
				m.clist.push_back( argv[i]);
			}
		}

	}
}
