/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "arena.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
#include <rg3DRom.h>

std::map <const std::string, int (*)( const gmCCArg &)> phArena::m_ccmd;

extern "C" gmThing *spw_phArena()
{
        return new phArena();
}

extern "C" int del_phArena( gmThing *t)
{
	delete t;
	return 0;
}

extern "C" int gm_main( const std::string *key, GameEngine *ge,
	std::string *emsg)
{
	ge->m_thing_manager->add_to_spawn_list( "phArena", spw_phArena);
	ge->m_thing_manager->add_to_delete_list( "phArena", del_phArena);
	return 0;
}

phArena::phArena()
{
	dsplst_ice_surface = 0;
	dsplst_boards = 0;

	m_ice_color = rgVector( 1, 1, 1);
	m_light_level = rgVector( 0.9, 0.9, 0.9);
	m_ice_alpha = 0.75;
	m_reflections = true;

	if( !(m_ccmd.empty())) return;
	add_console_command( "ice_color", phArena_ice_color, m_ccmd);
	add_console_command( "light_level", phArena_light_level, m_ccmd);
	add_console_command( "ice_alpha", phArena_ice_alpha, m_ccmd);
	add_console_command( "reflections", phArena_reflections, m_ccmd);
	add_console_command( "prepare_for_rendering",
		phArena_prepare_for_rendering, m_ccmd);
	add_console_command( "add_opaque", phArena_add_opaque, m_ccmd);
	add_console_command( "remove_opaque", phArena_remove_opaque, m_ccmd);
}

int phArena::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end()) return gmModelThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phArena::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++) {
		tmp += p_prefix + phArena::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += gmModelThing::list_ccmd( p_prefix);
	return tmp;
}

int phArena_ice_color( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 3) {
		p.ge->rcout( "usage: ice_color <red> <green> <blue>\n"
			"NOTE: All three components are given as floating "
			"point numbers\nwith a range of [0, 1]\n");
		return GM_THING_ERR_PARAM;
	}

	((phArena *)(p.thing))->m_ice_color = gm_str_to_vector( args[0],
		args[1], args[2]);
        return 0;
}

int phArena_ice_alpha( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 1) {
		p.ge->rcout( "usage: ice_alpha <value>\n"
			"NOTE: \"value\" is a number in the range [0, 1]\n");
		return GM_THING_ERR_PARAM;
	}
	((phArena *)(p.thing))->m_ice_alpha = strtod( (args[0]).c_str(), 0);
        return 0;
}

int phArena_light_level( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 3) {
		p.ge->rcout( "usage: light_level <red> <green> <blue>\n"
			"NOTE: All three components are given as floating "
			"point numbers\nwith a range of [0, 1]\n");
		return GM_THING_ERR_PARAM;
	}
	((phArena *)(p.thing))->m_light_level =
		gm_str_to_vector( args[0], args[1], args[2]);
	return 0;
}

int phArena_reflections( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	gm_set_state( ((phArena *)(p.thing))->m_reflections, args[0]);
	return 0;
}

int phArena_add_opaque( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	ThingControlBlock *tcb = p.ge->m_thing_manager->get( args[0]);
	if( !tcb) {
		p.ge->rcout( "No such thing \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	if( !(((phArena *)(p.thing))->add_opaque( tcb)))
		return GM_THING_ERR_GENERIC;
	return 0;
}

int phArena_remove_opaque( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	ThingControlBlock *tcb = p.ge->m_thing_manager->get( args[0]);
	if( !tcb) {
		p.ge->rcout( "No such thing \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	((phArena *)(p.thing))->remove_opaque( tcb);
	return 0;
}

int phArena_prepare_for_rendering( const gmCCArg &p)
{
	phArena *a = (phArena *)(p.thing);
	if( !(a->model())) {
		p.ge->rcout( "cannot prepare to render (no model)\n");
		return -1;
	}

	(a->model())->transform_skel_frame( a->m_draw_context);
	
	{	// For speed, render trusses and flags together in their own
		// display lists (keeping them seperate from the rest of the
		// arena).
		glNewList( a->create_display_list( "truss_flags"), GL_COMPILE);
		a->render_face_group( "truss", true, true);
		a->render_face_group( "flags", true, true);
		glEndList();
	}

	{	// Seats should be rendered in their own display list. Also,
		// include the arena box along with other parts of the stands.
		glNewList( a->create_display_list( "seats"), GL_COMPILE);
		a->render_face_group( "stands_floor", true, true);
		a->render_face_group( "seats", true, true);
		a->render_face_group( "arena_box", true, true);
		a->render_face_group( "floor", true, true);
		glEndList();
	}

	a->compile_face_list( "Ice Surface", a->dsplst_ice_surface);

	{ // Prepare the boards for rendering
		const char *tmp[] = { "Dasher Board Top", "Boards",
			"Top of boards (red part)"};
		a->compile_face_list( tmp, 3, a->dsplst_boards);
	}
	p.ge->rcout( "arena ready for rendering\n");
	return 0;
}

phArena::~phArena()
{
	clear();
}

void phArena::clear()
{
	clear_display_list( dsplst_ice_surface);
	clear_display_list( dsplst_boards);
}

void phArena::compile_face_list( const char *p_name, GLuint &id)
{
	const char *tmp[] = { p_name};
	compile_face_list( tmp, 1, id);
}

void phArena::compile_face_list( const char *p_name[], const int p_count,
	GLuint &id)
{
	if( id) { glDeleteLists( id, 1); id = 0;}
	id = glGenLists( 1);
	glNewList( id, GL_COMPILE);
	for( int i = 0; i < p_count; i++) {
		int count;
		const int *flist = (model())->get_faces_in_group(
			(model())->get_group_id( p_name[i]), count);
		if( flist) (model())->render_faces( true, true, flist, count);
	}
	glEndList();
}

void phArena::clear_display_list( GLuint &id)
{
	if( 0 == id) return;
	glDeleteLists( id, 1);
	id = 0;
}

void phArena::draw_boards() const
{
	if( dsplst_boards) glCallList( dsplst_boards);
}

void phArena::draw_glass()
{
	if( !(model())) return;
	gmCamera *c = (ge())->gfx_engine->current_camera();
	(model())->render_bsp_faces( m_draw_context, c->cs.p, true, true,
		0, 0);
}

// Description: The arena's "process" function handles rendering for everything
//	in the arena. This "process" function will only be called when
//	graphics need to be displayed. Therefore, features such as a
//	"jumbo-tron" (which requires processing beyond graphic updates)
//	should be created as a separate "thing" (not part of the arena class)
//	so that it can get its processing done as it likes. If a "thing" wants
//	to be rendered, then it must register with the arena, or a "thing"
//	which is already registered with the arena.
void phArena::process( ThingManager *tm)
{
	{ // Must clear colour buffer if wireframe rendering is on.
		GLint front_back[2];
		glGetIntegerv( GL_POLYGON_MODE, front_back);
		if(	(GL_LINE == front_back[0]) ||
			(GL_LINE == front_back[1]))
			glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		else glClear( GL_DEPTH_BUFFER_BIT);
	}
	

	// Render reflections
	// FIXME: Write an efficient (and correct) reflection algorithm.
	if( m_reflections) {
		glPushMatrix();
		glScalef( 1, -1, 1);
		glCullFace( GL_FRONT);

		// Boards
		glColor3( m_light_level);
		glDisable( GL_LIGHTING);
		draw_boards();
		glCallList( get_display_list( "truss_flags"));
		glCallList( get_display_list( "seats"));

		// "things"
		glEnable( GL_LIGHTING);
		glColor3f( 1, 1, 1);
		gmRenderArgs tmp;
		render_all_opaque( &tmp);

		// Glass
		glDisable( GL_LIGHTING);
		draw_glass();
		glClear( GL_DEPTH_BUFFER_BIT);

		glCullFace( GL_BACK);
		glPopMatrix();
	}

	// Render ice-surface
	glEnable( GL_BLEND);
	glColor4f( m_ice_color.x, m_ice_color.y, m_ice_color.z, m_ice_alpha);
	draw_ice_surface();
	glDisable( GL_BLEND);

	// Render remainder of arena
	// boards
	glColor3( m_light_level);
	draw_boards();
	glCallList( get_display_list( "truss_flags"));
	glCallList( get_display_list( "seats"));


	// Render all things with shadowing and no transparent parts.
	glEnable( GL_LIGHTING);
	glColor3f( 1, 1, 1);
	gmRenderArgs ra( 0, true);
	render_all_opaque( &ra);

	// Render all things again, only this time, only the transparent or
	// translucent parts are rendered.
	// FIXME: The "things" still have to be sorted to allow the rendering
	//	to wrok properly. For now, the problem is being ignored.
	glEnable( GL_BLEND);
	ra.render_opaque = false;
	ra.render_transparent = true;
	render_all_opaque( &ra);

	// Glass
	glDisable( GL_LIGHTING);
	draw_glass();

}

void phArena::render_all_opaque( gmRenderArgs *ra)
{
	std::map <ThingControlBlock *, void *> :: const_iterator iter =
                m_opaque.begin();
	std::vector <ThingControlBlock *> rm_list;
	for( ; iter != m_opaque.end(); iter++) {
		if( !((iter->first)->is_free()))
			(iter->first)->m_thing->draw( ra);
		else rm_list.push_back( iter->first);
	}
	while( !(rm_list.empty())) {
		remove_opaque( rm_list[0]);
		rm_list.erase( rm_list.begin());
	}
}

bool phArena::add_opaque( ThingControlBlock *tcb)
{
	typedef std::pair <ThingControlBlock *, void *> TPPair;
	return ((m_opaque.insert( TPPair( tcb, 0))).second);
}

void phArena::remove_opaque( ThingControlBlock *tcb)
{
	std::map <ThingControlBlock *, void *> :: iterator iter =
		m_opaque.find( tcb);
	if( iter == m_opaque.end()) return;
	m_opaque.erase( iter);
}
