/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#ifndef ENGINE_GAME_H
#define ENGINE_GAME_H

#include <pthread.h>
#include <iostream>
#include <map>
#include <vector>
#include <string>
#include "main_header.h"

#define BOOTUP_FILE_NAME	"rph_bootup.cfg"
#define STARTUP_FILE_NAME	"rph_startup.cfg"

#define MAX_BATCH_LINE_LENGTH	64000

class MainData;
class rgTimer;
class rgSocket;
class GraphicsEngine;
class PhysicsEngine;
class sk3DRomModel;
class ThingManager;
class GameConsole;
class rgVector;

class gmInputAction
{
public:
	std::vector <std::string> m_up;
	std::vector <std::string> m_down;
};

class GameEngine
{
public:
	// XXX: Physics
	PhysicsEngine *phy_engine;

	// XXX: Graphics
	GraphicsEngine *gfx_engine;

	// XXX: Things
	ThingManager *m_thing_manager;


	GameEngine();
	~GameEngine();

	int load_startup_file( const char *fname);
	int process();

	// Description: Returns the number of game clocks passed since the
	//	game was started. The game clock always counts up. Therefore,
	//	there is a risk of the clock "wrapping around" after it
	//	exceeds the precision of the data type storing the clock ticks.
	//	However, the least precise data type is a 64-bit integer which
	//	when running at 1000 game clocks per second, should take about
	//	500,000,000 years before a "wrap-around" occurs.
	long long game_clock() const { return m_game_clock;}

	// Description: Returns the number of game clocks per second.
	long long game_clock_hz() const { return m_game_clock_hz;}

	// Description: Returns the number of milliseconds in a single game
	//	clock tick.
	long long game_clock_tick() const {
		return (long long)(1000.0f * m_game_clock_tick);}

	void clear();

	//int load_arena( const char *fname);

	ThingManager *thing_manager() { return m_thing_manager;}

	void rcout( const char *msg) const;
	void rcout( const std::string &msg) const;

	bool add_console_command( const std::string &key,
		int (*fn)(const std::vector <std::string> &, GameEngine &));
	int exec_key( const std::string &key, const std::string &up_down);

	int process_message( const std::string &msg);

	bool engine_ok() const {
		if( !m_engine_started && m_shutdown) return false;
		return true;
	}	
	int startup();

	// FIXME: "mdata" should be private.
	MainData *mdata;
private:

	bool m_engine_started;
	bool m_shutdown;

	// XXX: "Things"
	int create_thing_manager();

	// XXX: Networking
	bool m_use_console_server;
	GameConsole *m_console;

	// XXX: Timing
	double m_game_clock_scalar;
	double m_game_clock_tick;
	long long m_game_clock_hz;
	long long m_game_clock;
	rgTimer *m_game_timer;
	long long game_clocks_pending();
	
	// XXX: Input
	std::map <const std::string, gmInputAction> m_key_binding;
	gmInputAction *binding( const std::string &key);
	int add_binding( const std::string &key,
		const std::vector <std::string> &args,
		const std::string &up_down, bool to_front);
	int unbind( const std::string &key);
	std::string dump_bound_keys();
	std::string dump_binding( const std::string &key);

	// XXX: Console messages
	bool m_skip_batch_lines;
	std::map <const std::string, const std::string> m_defines;
	std::map <const std::string,
		int (*)(const std::vector <std::string> &, GameEngine &)>
		m_commands;
	std::string m_pending_console_cmd;
	void check_for_console_messages();
	int gen_arg_list( const std::string msg,
		std::vector <std::string> &arg, std::string &emsg) const;
	int add_definition( const std::string &key, const std::string &txt);
	int remove_definition( const std::string &key);
	std::string get_definition( const std::string &key, int &err) const;

	int m_batch_limit;
	int process_batch_file( const std::string &fname, bool err_fatal);


	#define TMP_ARGS const std::vector <std::string> &args, GameEngine &ge
	friend int ccmd_add_bounding_thing( TMP_ARGS);
	friend int ccmd_add_camera( TMP_ARGS);
	friend int ccmd_batch( TMP_ARGS);
	friend int ccmd_bind( TMP_ARGS);
	friend int ccmd_bind_front( TMP_ARGS);
	friend int ccmd_bindings( TMP_ARGS);
	friend int ccmd_build_sector_map( TMP_ARGS);
	friend int ccmd_clear_color( TMP_ARGS);
	friend int ccmd_clear_sector_map( TMP_ARGS);
	friend int ccmd_color_material( TMP_ARGS);
	friend int ccmd_compile_model_alpha_faces( TMP_ARGS);
	friend int ccmd_cross( TMP_ARGS);
	friend int ccmd_cullface( TMP_ARGS);
	friend int ccmd_define( TMP_ARGS);
	friend int ccmd_depth_test( TMP_ARGS);
	friend int ccmd_do( TMP_ARGS);
	friend int ccmd_dot( TMP_ARGS);
	friend int ccmd_draw_sectors( TMP_ARGS);
	friend int ccmd_dump_all_things( TMP_ARGS);
	friend int ccmd_dump_camera( TMP_ARGS);
	friend int ccmd_dump_model( TMP_ARGS);
	friend int ccmd_dump_sector_map( TMP_ARGS);
	friend int ccmd_dump_thing( TMP_ARGS);
	friend int ccmd_echo( TMP_ARGS);
	friend int ccmd_endif( TMP_ARGS);
	friend int ccmd_game_clock( TMP_ARGS);
	friend int ccmd_game_clock_scalar( TMP_ARGS);
	friend int ccmd_gfx_engine( TMP_ARGS);
	friend int ccmd_gravity( TMP_ARGS);
	friend int ccmd_ifdef( TMP_ARGS);
	friend int ccmd_ifndef( TMP_ARGS);
	friend int ccmd_kill( TMP_ARGS);
	friend int ccmd_list_cameras( TMP_ARGS);
	friend int ccmd_list_commands( TMP_ARGS);
	friend int ccmd_list_models( TMP_ARGS);
	friend int ccmd_list_spawnable( TMP_ARGS);
	friend int ccmd_load_model( TMP_ARGS);
	friend int ccmd_load_model_textures( TMP_ARGS);
	friend int ccmd_load_module( TMP_ARGS);
	friend int ccmd_load_texture( TMP_ARGS);
	friend int ccmd_normalize( TMP_ARGS);
	friend int ccmd_perspective( TMP_ARGS);
	friend int ccmd_ping( TMP_ARGS);
	friend int ccmd_pwd( TMP_ARGS);
	friend int ccmd_quit( TMP_ARGS);
	friend int ccmd_remove_camera( TMP_ARGS);
	friend int ccmd_remove_model( TMP_ARGS);
	friend int ccmd_remove_thing( TMP_ARGS);
	friend int ccmd_set_alt_texture( TMP_ARGS);
	friend int ccmd_set_camera_position( TMP_ARGS);
	friend int ccmd_set_current_camera( TMP_ARGS);
	friend int ccmd_set_game_clock_hz( TMP_ARGS);
	friend int ccmd_set_root_renderer( TMP_ARGS);
	friend int ccmd_set_screen( TMP_ARGS);
	friend int ccmd_set_server_port_range( TMP_ARGS);
	friend int ccmd_set_texture_color_key( TMP_ARGS);
	friend int ccmd_spawn( TMP_ARGS);
	friend int ccmd_thing_list_check( TMP_ARGS);
	friend int ccmd_thing_manager( TMP_ARGS);
	friend int ccmd_unbind( TMP_ARGS);
	friend int ccmd_undef( TMP_ARGS);
	friend int ccmd_use_console_server( TMP_ARGS);
	friend int ccmd_wireframe( TMP_ARGS);
	#undef TMP_ARGS

};

#define MAX_CONSOLE_MESSAGE_LENGTH	8192
class ConsoleMessage
{
public:
	int type;
	int length;
	char msg[MAX_CONSOLE_MESSAGE_LENGTH];
	bool console_lock;	// Console server is reading/writing
	bool game_lock;		// Game engine is reading/writing

	ConsoleMessage() {
		type = 0;
		length = 0;
		msg[0] = '\0';
		msg[MAX_CONSOLE_MESSAGE_LENGTH - 1] = '\0';
		console_lock = true;
		game_lock = false;
	}

	void print_message( std::ostream &o);
};

class GameConsole
{
public:
	int m_first_port;
	int m_last_port;

	rgSocket *m_sock;

	GameConsole( GameEngine *ge);
	~GameConsole();

	int start_server_thread();

protected:
	bool m_server_started;
	pthread_t m_server_thread;
	GameEngine *ge;

	friend int hlp_get_message( GameConsole *gc);
	friend void *console_server_thread( void *ptr);
};

bool gm_set_state( bool &des, const std::string &s);
bool gm_set_state( const std::string &s);

// FIXME: "gm_to_string" should be probably be "templated".
std::string gm_to_string( const void *p);
std::string gm_to_string( const int i);
std::string gm_to_string( const float f);
std::string gm_to_string( const double f);
std::string gm_to_string( const rgVector &v);
rgVector gm_str_to_vector( std::string x, std::string y, std::string z);

#endif
