/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "engine_game.h"
#include "engine_graphics.h"
#include "engine_physics.h"
#include "things/things.h"
#include "vector"
#include <rgNetwork.h>
#include <rgOpenGL.h>
#include <rgPort.h>
#include <rgStandard.h>
#include <rgError.h>
#include <rg3DRom.h>
#include <fstream>
#include <GL/glu.h>
#include <sstream>

bool hlp_bad_def_char( const char c, std::string &emsg)
{
	switch( c) {
	case '\\':
	case '"':
	case '$': return true;
	default: return false;
	}
}

int GameEngine::add_definition( const std::string &key, const std::string &txt)
{
	typedef std::pair <const std::string, const std::string> SSPair;
	if( false == (m_defines.insert( SSPair( key, txt)).second)) return -1;
	return 0;
}

int GameEngine::remove_definition( const std::string &key)
{
	std::map <const std::string, const std::string> :: iterator iter =
		m_defines.find( key);
	if( iter == m_defines.end()) return -1;
	m_defines.erase( iter);
	return 0;
}

std::string GameEngine::get_definition( const std::string &key, int &err) const
{
	std::map <const std::string, const std::string> :: const_iterator iter
		= m_defines.find( key);
	if( iter == m_defines.end()) {
		err = -1;
		return std::string();
	}
	err = 0;
	return (iter->second);
}

int GameEngine::gen_arg_list( const std::string msg,
	std::vector <std::string> &arg, std::string &emsg) const
{
	arg.clear();
	bool definition = false;
	bool escape = false;
	bool in_quote = false;
	std::string tmp;
	for( int i = 0; i < int( msg.length()); i++) {
		if( escape) {
			switch( msg[i]) {
			case 'n': tmp += '\n'; break;
			case '0': tmp += '\0'; break;
			case 't': tmp += '\t'; break;
			case 'r': tmp += '\r'; break;
			case 'b': tmp += '\b'; break;
			case ' ':
			case '\t':
			case '\n':
				if( tmp.length() > 0) {
					if( in_quote) tmp += msg[i];
					else {
						arg.push_back( tmp);
						tmp.clear();
					}
				}
				break;
			default: tmp += msg[i]; break;
			}
			escape = false;

		} else if( in_quote) {
			switch( msg[i]) {
			case '"': in_quote = false; break;
			case '\\': escape = true; break;
			default: tmp += msg[i]; break;
			}

		} else {
			if( definition && hlp_bad_def_char( msg[i], emsg))
				return -1;
			switch( msg[i]) {
			case '\\': escape = true; break;
			case '$':
				if( !m_skip_batch_lines) definition = true;
				break;
			case '"': in_quote = true; break;

			case ' ':
			case '\t':
			case '\n':
				if( definition) {
					int err;
					const std::string txt =
						get_definition( tmp, err);
					if( err) {
						emsg = "No such definition \""
							+ tmp + "\"\n";
						return -1;
					}
					tmp = txt;
					definition = false;
				}
				if( tmp.length() > 0) {
					arg.push_back( tmp);
					tmp = std::string();
				}
				break;
	
			default:
				tmp += msg[i];
				break;
			}
		}
	}
	if( tmp.length() > 0) arg.push_back( tmp);
	return arg.size();
}

void GameEngine::rcout( const char *msg) const
{
	if( !(m_console->m_sock->connected())) {
		std::cout << msg;
		std::cout.flush();

	} else m_console->m_sock->send( msg, strlen( msg));
}

void GameEngine::rcout( const std::string &msg) const
{
	if( !(m_console->m_sock->connected())) {
		std::cout << msg;
		std::cout.flush();

	} else m_console->m_sock->send( msg.c_str(), msg.length());
}

bool gm_set_state( bool &des, const std::string &s)
{
	if( ("1" == s) || ("on" == s) || ("true" == s)) des = true;
	else if( ("0" == s) || ("off" == s) || ("false" == s)) des = false;
	else if( "toggle" == s) des = !des;
	return des;
}

bool gm_set_state( const std::string &s)
{
	bool b = false;
	return gm_set_state( b, s);
}

bool hlp_good_bool_cmd( const char *cmd,
	const std::vector <std::string> &args,
	const GameEngine *ge)
{
	if( args.size() != 2) {
		std::string s( "usage: ");
		s += cmd;
		s += " <0/1/on/off/toggle>\n";
		ge->rcout( s.c_str());
		return false;
	}
	return true;	
}

int GameEngine::process_message( const std::string &msg)
{
	std::vector <std::string> args;
	std::string emsg;
	const int err = gen_arg_list( msg, args, emsg);
	if( 0 == err) return 0;
	if( err < 0) {
		rcout( emsg);
		return -1;
	}

	if( '#' == ((args[0])[0])) return 0;
	if( m_skip_batch_lines && (args[0] != "endif")) return 0;
	std::map <	const std::string,
		int (*)( const std::vector <std::string> &, GameEngine &)>
		:: const_iterator iter;
	iter = m_commands.find( args[0]);
	if( iter == m_commands.end()) {
		rcout( "no such command \"" + args[0] + "\"\n");
		return -2;
	
	}
	return ((iter->second)( args, *this));
}

int GameEngine::process_batch_file( const std::string &fname, bool err_fatal)
{
	std::ifstream src( fname.c_str());
	if( !(src.is_open())) return -1;
	rgTempArray <char> buf( MAX_BATCH_LINE_LENGTH);
	if( !(buf.data)) return -2;
	int line_number = 0;
	while( true) {
		int cmd_start_line = line_number + 1;
		int c = 0;
		do {
			line_number++;
			src.getline( buf.data + c,
				(MAX_BATCH_LINE_LENGTH - 1) - c);
			// XXX: Minus one cause '\n' not read.
			c += src.gcount() - 1;
		} while( '\\' == buf.data[c - 1]);
		buf.data[ MAX_BATCH_LINE_LENGTH - 1] = '\0';
		if( "exit_batch_file" == std::string( buf.data)) break;
		const int err = process_message( std::string( buf.data));
		if( err && err_fatal) return cmd_start_line;
		if( src.eof()) break;
		if( src.fail()) return cmd_start_line;
	}
	return 0;
}

std::string gm_to_string( const void *p)
{
	std::ostringstream des;
	des << p;
	return des.str();
}

std::string gm_to_string( const int i)
{
	std::ostringstream des;
	des << i;
	return des.str();
}

std::string gm_to_string( const float f)
{
	std::ostringstream des;
	des << f;
	return des.str();
}

std::string gm_to_string( const double f)
{
	std::ostringstream des;
	des << f;
	return des.str();
}

std::string gm_to_string( const rgVector &v)
{
	std::ostringstream des;
	des << v;
	return des.str();
}

rgVector gm_str_to_vector( std::string x, std::string y, std::string z)
{
	return rgVector(
		strtod( x.c_str(), 0),
		strtod( y.c_str(), 0),
		strtod( z.c_str(), 0));
}

// XXX: "ccmd_" (Console CoMmanD).
#define TMP_ARGS const std::vector <std::string> &args, GameEngine &ge
int ccmd_ifdef( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: ifdef <definition>\n");
		return -1;
	}
	int err;
	ge.get_definition( args[1], err);
	if( err) ge.m_skip_batch_lines = true;
	else ge.m_skip_batch_lines = false;
	return 0;
}

int ccmd_ifndef( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: ifndef <definition>\n");
		return -1;
	}
	int err;
	ge.get_definition( args[1], err);
	if( err) ge.m_skip_batch_lines = false;
	else ge.m_skip_batch_lines = true;
	return 0;
}

int ccmd_endif( TMP_ARGS)
{
	ge.m_skip_batch_lines = false;
	return 0;
}

int ccmd_define( TMP_ARGS)
{
	if( args.size() < 2) {
		ge.rcout( "usage: define <definition name> [definition text] "
				"[definition text] [definition text] ...\n"
			"NOTE: Extra spaces/tabs are not preserved in the\n"
				"definition text, unless you use quotes (\"\")"
				".\n");
		return -1;
	}
	std::string tmp;
	if( args.size() >= 3) tmp += args[2];
	for( int i = 3; i < int( args.size()); i++)
		tmp += " " + args[i];
	int err;
	ge.get_definition( args[1], err);
	if( !err) {
		ge.rcout( "Redefinition of \"" + args[1] + "\"\n");
		return -2;
	}
	if( ge.add_definition( args[1], tmp)) {
		ge.rcout( "could not add definition \"" + args[1] + "\"\n");
		return -3;
	}
	return 0;
}

int ccmd_undef( TMP_ARGS)
{
	if( args.size() < 2) {
		ge.rcout( "usage: undef <definition> [definition] [definition]"
			" ...\n"
			"NOTE: No error message is generated if a given "
				"definition\ndoes not exist.\n");
		return -1;
	}
	for( int i = 1; i < int( args.size()); i++)
		ge.remove_definition( args[i]);
	return 0;
}

int ccmd_kill( TMP_ARGS)
{
	exit(0);
	return 0;	// <-- Statment never reached
}

int ccmd_quit( TMP_ARGS)
{
	ge.m_shutdown = true;
	return 0;
}

int ccmd_game_clock_scalar( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: game_clock_scalar <scalar>\n"
			"NOTE: \"scalar\" is used to speed up or slow down "
			"the game timing.\n"
			"A scalar value of 0.5 will make the game engine "
				"\"think\" that\n"
			"it only needs to process half of the scheduled "
			"frames, thus making it \n"
			"run half as fast (50\% of normal).\n"
			"Conversley, a scalar value of 2.0 will make the "
			"game engine run twice as fast.\n"
			"A scalar value of 0 will \"pause\" the game.\n"
			"(These are \"game\" frames, not \"graphics\" "
				"frames.)\n");
		return -1;
	}
	ge.m_game_clock_scalar = strtod( (args[1]).c_str(), 0);
	ge.rcout( "Game time scalar = " +
		gm_to_string( ge.m_game_clock_scalar) + "\n");
	return 0;
}

int ccmd_echo( TMP_ARGS)
{
	for( int i = 1; i < int( args.size()); i++) ge.rcout( args[i]);
	return 0;
}

int ccmd_unbind( TMP_ARGS)
{
	if( args.size() <= 1) {
		ge.rcout( "usage: unbind <which key> [which key] [which key] "
				"[which key]...\n"
			"NOTE: \"which key\" is the name of the key or the "
			"character equivalent.\n"
			"example 1: unbind \"c\"\n"
			"example 2: unbind \"right_arrow\"\n");
		return -1;
	}
	for( int i = 1;	i < int( args.size()); i++) {
		if( ge.unbind( args[i])) {
			ge.rcout( "Could not unbind key \"" + args[i] +
				"\"\n");
		}
	}
	return 0;
}

int ccmd_bindings( TMP_ARGS)
{
	ge.rcout( ge.dump_bound_keys());
	return 0;
}

int ccmd_bind_front( TMP_ARGS)
{
	if( args.size() <= 1) {
		ge.rcout( "usage: bind_front <which key> <up/down> [command] "
				"[arg 1] [arg 2] [arg 3] ...\n"
			"NOTE: \"which key\" is the name of the key or the "
			"character equivalent.\n"
			"\"up/down\" indicates whether the command has effect "
				"on a key press or release.\n"
			"Passing no arguments to \"bind_front\" will display "
				"the current bindings for the key\n"
			"example 1: bind_front \"c\" \"down\" list_commands\n"
			"example 2: bind_front \"j\" \"down\" do jump "
				"thg_player\n"
			"example 3: bind_front \"right_arrow\" \"down\" "
				"do move_right thg_player\n"
			"example 4: bind_front \"right_arrow\" \"up\" "
				"do stop_moving thg_player\n"
			"SEE ALSO: bindings, bind, unbind\n");
		return -1;

	} else if( 2 == args.size()) {
		ge.rcout( ge.dump_binding( args[1]));
		return 0;

	} else if( 3 == args.size()) {
		ge.rcout( "Nothing to bind, ignoring command\n");
		return 0;
	}

	std::vector <std::string> tmp;
	for( int i = 3; i < int( args.size()); i++) tmp.push_back( args[i]);
	if( ge.add_binding( args[1], tmp, args[2], true)) {
		ge.rcout( "Could not bind key \"" + args[1] + "\"\n");
		return -2;
	}
	return 0;
}

int ccmd_bind( TMP_ARGS)
{
	if( args.size() <= 1) {
		ge.rcout( "usage: bind <which key> <up/down> [command] [arg 1]"
				" [arg 2] [arg 3] ...\n"
			"NOTE: \"which key\" is the name of the key or the "
			"character equivalent.\n"
			"\"up/down\" indicates whether the command has effect "
				"on a key press or release.\n"
			"Passing no arguments to \"bind\" will display the "
				"current bindings for the key\n"
			"example 1: bind \"c\" \"down\" list_commands\n"
			"example 2: bind \"j\" \"down\" do jump thg_player\n"
			"example 3: bind \"right_arrow\" \"down\" "
				"do move_right thg_player\n"
			"example 4: bind \"right_arrow\" \"up\" "
				"do stop_moving thg_player\n"
			"SEE ALSO: bindings, bind_front, unbind\n");
		return -1;

	} else if( 2 == args.size()) {
		ge.rcout( ge.dump_binding( args[1]));
		return 0;

	} else if( 3 == args.size()) {
		ge.rcout( "Nothing to bind, ignoring command\n");
		return 0;
	}

	std::vector <std::string> tmp;
	for( int i = 3; i < int( args.size()); i++) tmp.push_back( args[i]);
	if( ge.add_binding( args[1], tmp, args[2], false)) {
		ge.rcout( "Could not bind key \"" + args[1] + "\"\n");
		return -2;
	}
	return 0;
}

int ccmd_ping( TMP_ARGS)
{
	ge.rcout( "pong\n");
	return 0;
}

int ccmd_add_bounding_thing( TMP_ARGS)
{
	const std::string more_info(
		"NOTE: \"key\" must be the indentifier of a "
		"\"gmModelThing\".\n"
		"The bounding cubes/prisims from the \"things\" "
		"model are added to the\n"
		"physics engine's sector map (used in collision "
		"detection).\n"
		"You can only add bounding models if no sector map "
		"has yet been built.\n"
		"Otherwise, you must delete the sector map first\n");

	if( args.size() < 2) {
		ge.rcout( "usage: add_bounding_model <key>\n" + more_info);
		return -1;
	}

	ThingControlBlock *tcb = ge.m_thing_manager->get( args[1]);
	if( !tcb) {
		ge.rcout( "No such thing \"" + args[1] + "\"\n");
		return -2;
	}

	if( !(tcb->m_thing->is_a( "gmModelThing"))) {
		ge.rcout( "Thing is not a type of \"gmModelThing\"\n" +
			more_info);
		return -3;
	}

	if( !(ge.phy_engine->m_sector_map.add_prisims(
		(gmModelThing *)(tcb->m_thing)))) {
		ge.rcout( "Could not add bounding thing\n" + more_info);
		return -4;
	}

	ge.rcout( "Added bounding thing \"" + args[1] + "\" to sector "
		"map\n");
	return 0;
}

int ccmd_build_sector_map( TMP_ARGS)
{
	if( args.size() != 13) {
		ge.rcout( "usage: build_sector_map <a> <b> <c> <d>\n"
			"NOTE: All four parameters (a, b, c, d) are vectors.\n"
			"These specify the starting dimensions of the quad "
				"tree sector map.\n"
			"They form a counter clockwise quad, with \"a\" being "
				"the top right corner.\n"
			"The quad is axis aligned.\n");
		return -1;
	}
	if( !(ge.phy_engine->m_sector_map.build_map(
		gm_str_to_vector( args[1], args[2], args[3]),
		gm_str_to_vector( args[4], args[5], args[6]),
		gm_str_to_vector( args[7], args[8], args[9]),
		gm_str_to_vector( args[10], args[11], args[12])))) {
		ge.rcout( "Could not build sector map\n");
		return -2;
	}
	ge.rcout( "Sector map built\n");	
	return 0;
}

int ccmd_clear_sector_map( TMP_ARGS)
{
	ge.phy_engine->m_sector_map.clear();
	ge.rcout( "Sector map cleared\n");
	return 0;
}

int ccmd_do( TMP_ARGS)
{
	if( args.size() < 3) {
		ge.rcout( "usage: do <command> <key> [<arg>...<arg>...]\n"
			"NOTE: \"command\" is the command to perform.\n"
			"\"key\" is the thing to perform the command on.\n"
			"Additional (optional) arguments can be passed.\n");
		return -1;
	}
	ThingControlBlock *tcb = ge.m_thing_manager->get( args[2]);
	if( !tcb) {
		ge.rcout( "No such thing \"" + args[2] + "\"\n");
		return -2;
	}
	std::vector <std::string> tmp;
	for( int i = 3; i < int( args.size()); i++) tmp.push_back( args[i]);
	gmCCArg tmp_arg( tcb->m_thing, &ge, &tmp);
	const int err = tcb->m_thing->call_ccmd( args[1], tmp_arg);
	if( err) {
		ge.rcout( "Error " + gm_to_string( err) +
			" while executing console command \"" + args[1] +
			"\" for thing \"" + args[2] + "\"\n");
		return -3;
	}
	return 0;
}

int ccmd_batch( TMP_ARGS)
{
	bool err_fatal = true;
	switch( args.size()) {
	case 2: break;
	case 3:
		if( "no" == args[2]) err_fatal = false;
		break;

	default:
		ge.rcout( "usage: batch <file> [no]\n"
			"NOTE: \"file\" is the batch file to load (you can "
			"also specify a path).\n"
			"\"no\" is optional and indicates that errors in a "
			"batch file are\nnot fatal, otherwise, without \"no\""
			", errors are fatal.\n");
		return -1;
	}

	// FIXME: Make sure recursive batch files aren't loaded. Lock the
	//	file so it cannot be read again recursively. For now, an
	//	arbitrary batch recursion limit is impossed (not ideal).
	if( !(ge.m_batch_limit)) return -1000;
	ge.m_batch_limit--;
	const int err = ge.process_batch_file( args[1], err_fatal);
	ge.m_batch_limit++;
	if( err < 0) {
		ge.rcout( "Error " + gm_to_string( err) +
			" while processing batch file \"" + args[1] + "\"\n");
		return -2;

	} else if( err > 0) {
		ge.rcout( "Error on line " + gm_to_string( err) +
			" while processing batch file \"" + args[1] + "\"\n");
		return -3;
	}
	return 0;
}

int ccmd_dump_sector_map( TMP_ARGS)
{
	ge.rcout( ge.phy_engine->m_sector_map.dump());
	return 0;
}

int ccmd_set_screen( TMP_ARGS)
{
	if( args.size() != 5) {
		ge.rcout( "usage: set_screen <width> <height> <bpp> "
			"<zbuffer depth>\n"
			"NOTE: \"bpp\" = Bits Per Pixel of colour.\n"
			"\"zbuffer depth\" is given in bits per pixel.\n");
		return -1;
	}
	const int width = atoi( (args[1]).c_str());
	const int height = atoi( (args[2]).c_str());
	const int bpp = atoi( (args[3]).c_str());
	const int zbuffer = atoi( (args[4]).c_str());
	if( width <= 0) {
		ge.rcout( "Invalid width = \"" + args[1] + "\"\n");
		return -2;
	}
	if( height <= 0) {
		ge.rcout( "Invalid height = \"" + args[2] + "\"\n");
		return -3;
	}
	
	ge.mdata->m_video_mode.push_back( gmVideoInfo( width, height, bpp,
		zbuffer));
	return 0;
}

int ccmd_set_game_clock_hz( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: set_game_clock_hz <frames per second>\n");
		return -1;
	}
	if( ge.m_engine_started) {
		ge.rcout( "\"" + args[0] + "\" only valid in bootup file.\n");
		return -2;
	}
	const int hz = atoi( (args[1]).c_str());
	if( hz <= 0) {
		ge.rcout( args[1] + " is an invalid frequency\n");
		return -3;
	}
	ge.m_game_clock_hz = hz;
	ge.rcout( "Game clock hz: " + gm_to_string( int( ge.m_game_clock_hz)) +
		"\n");
	return 0;
}

int ccmd_set_server_port_range( TMP_ARGS)
{
	if( args.size() != 3) {
		ge.rcout( "usage: set_server_port_range <first> <last>\n");
		return -1;
	}
	if( ge.m_engine_started) {
		ge.rcout( "\"" + args[0] + "\" only valid in bootup file.\n");
		return -2;
	}
	ge.m_console->m_first_port = atoi( (args[1]).c_str());
	ge.m_console->m_last_port = atoi( (args[2]).c_str());
	ge.rcout( "Server port range: " +
		gm_to_string( ge.m_console->m_first_port) + " to " +
		gm_to_string( ge.m_console->m_last_port) + "\n");
	return 0;
}

int ccmd_dump_camera( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: dump_camera <key>\n");
		return -1;
	}
	ge.rcout( ge.gfx_engine->dump_camera( args[1]));
	return 0;
}

int ccmd_remove_camera( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: remove_camera <key>\n"
			"NOTE: Camera's that have owners cannot be removed.\n"
			);
		return -1;
	}
	const int err = ge.gfx_engine->remove_camera( args[1]);
	if( err) {
		ge.rcout( "Error " + gm_to_string( err) +
			"while removing camera with key = \"" + args[1] +
			"\"\n");
		return -2;
	}
	ge.rcout( "Camera \"" + args[1] + "\" removed\n");
	return 0;
}

int ccmd_list_cameras( TMP_ARGS)
{
	ge.rcout( ge.gfx_engine->list_cameras());
	return 0;
}

int ccmd_add_camera( TMP_ARGS)
{
	if( (args.size() != 8) && (args.size() != 11)) {
		ge.rcout( "usage: add_camera <key> <pX> <pY> <pZ> "
			"<lX> <lY> <lZ> [<uX> <uY> <uZ>]\n"
			"NOTE: \"key\" uniquely identifies the camera.\n"
			"The camera's position is given by ( pX, pY, pZ)\n"
			"The camera's \"look at\" point is given by "
				"( lX, lY, lZ)\n"
			"The camera's \"up\" vector is given by "
				"( uX, uY, uZ)\n"
			"The \"up\" vector specification is optional\n");
		return -1;
	}
	const rgVector p( gm_str_to_vector( args[2], args[3], args[4]));
	const rgVector look( gm_str_to_vector( args[5], args[6], args[7]));
	int err;
	if( 8 == args.size()) {
		err = ge.gfx_engine->add_camera( args[1], p, look);

	} else {
		rgVector up( gm_str_to_vector( args[8], args[9], args[10]));
		err = ge.gfx_engine->add_camera( args[1], p, look, up);
	}
	if( err) {
		ge.rcout( "Error " + gm_to_string( err) +
			" while adding camera with key = \"" + args[1] +
			"\"\n");
		return -2;
	}
	ge.rcout( "Camera \"" + args[1] + "\" added\n");
	return 0;
}

int ccmd_set_current_camera( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: set_current_camera <key>\n");
		return -1;
	}
	if( !(ge.gfx_engine->set_current_camera( args[1]))) {
		ge.rcout( "Could not make current camera with key = \"" +
			args[1] + "\"\n");
		return -2;
	}
	return 0;
}

int ccmd_set_camera_position( TMP_ARGS)
{
	if( args.size() != 5) {
		ge.rcout( "usage: set_camera_position <key> <x> <y> <z>\n");
		return -1;
	}
	gmCamera *c = ge.gfx_engine->get_camera( args[1]);
	if( !c) {
		ge.rcout( "Could not get camera with key = \"" +
			args[1] + "\"\n");
		return -2;
	}
	c->cs.p = gm_str_to_vector( args[2], args[3], args[4]);
	return 0;
}

int ccmd_cross( TMP_ARGS)
{
	if( args.size() != 7) {
		ge.rcout( "usage: cross <a> <b> <c> <u> <v> <w>\n"
			"NOTE: Takes cross product of vectors (a, b, c) and\n"
			"(u, v, w)\n");
		return -1;
	}
	const rgVector a( gm_str_to_vector( args[1], args[2], args[3]));
	const rgVector b( gm_str_to_vector( args[4], args[5], args[6]));
	ge.rcout( "( " + args[1] + ", " + args[2] + ", " + args[3] +
		") cross ( " + args[4] + ", " + args[5] + ", " + args[6] +
		") = " + gm_to_string( a.cross( b)) + "\n");
	return 0;
}

int ccmd_dot( TMP_ARGS)
{
	if( args.size() != 7) {
		ge.rcout( "usage: dot <a> <b> <c> <u> <v> <w>\n"
			"NOTE: Takes dot product of vectors (a, b, c) and\n"
			"(u, v, w)\n");
		return -1;
	}
	const rgVector a( gm_str_to_vector( args[1], args[2], args[3]));
	const rgVector b( gm_str_to_vector( args[4], args[5], args[6]));
	ge.rcout( "( " + args[1] + ", " + args[2] + ", " + args[3] +
		") dot ( " + args[4] + ", " + args[5] + ", " + args[6] +
		") = " + gm_to_string( a.dot( b)) + "\n");
	return 0;
}

int ccmd_normalize( TMP_ARGS)
{
	if( args.size() != 4) {
		ge.rcout( "usage: normalize <x> <y> <z>\n"
			"NOTE: Normalizes the vector (x, y, z)\n");
		return -1;
	}
	rgVector a( gm_str_to_vector( args[1], args[2], args[3]));
	ge.rcout( "( " + args[1] + ", " + args[2] + ", " + args[3] +
		") normalized = " + gm_to_string( a.normalize()) +
		"\n");
	return 0;
}

int ccmd_gravity( TMP_ARGS)
{
	switch( args.size()) {
	case 1:	ge.rcout( "gravity = " + gm_to_string(
			ge.phy_engine->m_gravity *
				ge.phy_engine->gravity_frame_scalar()) +
			" m/s\n");
		break;

	case 4: {
		// XXX: The gravity "m_gravity" is stored internally as
		//	meters per frame (since physics occurs on a per frame
		//	basis).
			const float k = 1.0f /
				ge.phy_engine->gravity_frame_scalar();
			rgVector &gravity = ge.phy_engine->m_gravity;
			gravity.x = k * strtod( (args[1]).c_str(), 0);
			gravity.y = k * strtod( (args[2]).c_str(), 0);
			gravity.z = k * strtod( (args[3]).c_str(), 0);
		}
		break;

	default: ge.rcout( "usage: gravity <X> <Y> <Z>\n"
			"NOTE: Gravity is a vector quantity in meters per "
			"second. Also,\ntype \"gravity\" without parameters "
			"to display the current gravity\n");
		return -1;
	}
	return 0;
}

int ccmd_clear_color( TMP_ARGS)
{
	if( args.size() != 4) {
		ge.rcout( "usage: clear_color <red> <green> <blue>\n"
			"NOTE: Use numbers in the range [0, 1] for "
			"all colour components\n");
		return -1;
	}
	glClearColor(
		strtod( (args[1]).c_str(), 0),
		strtod( (args[2]).c_str(), 0),
		strtod( (args[3]).c_str(), 0),
		1);
	return 0;
}

int ccmd_dump_thing( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: dump_thing <key>\n");
		/*
		ge.rcout( "usage: dump_thing <id number/key>\n"
			"NOTE: The argument to \"dump_thing\" can be an "
			"index number of a key value.\n"
			"If a given key is a number, then both the \"thing\" "
			"with a\nmatching number and a matching key are "
			"displayed\n");
		*/
		return -1;
	}
	/*
	const char *tmp = (args[1]).c_str();
	char *endptr;
	const int id = strtod( tmp, &endptr);
	std::string des;
	if( tmp != endtpr) {
		des = ge.m_thing_manager.
	}
	*/
	ge.rcout( ge.m_thing_manager->dump_thing( args[1]));
	return 0;
}

int ccmd_pwd( TMP_ARGS)
{
	const int len = 300;
	char buf[len];
	if( !rgport_getcwd( buf, len)) {
		ge.rcout( "Could not get current working directory\n");
		return -1;
	}
	ge.rcout( std::string( buf) + "\n");
	return 0;
}

int ccmd_dump_all_things( TMP_ARGS)
{
	if( 2 == args.size()) {
		const int len = 300;
		char buf[len];
		if( !(rgport_getcwd( buf, len)))
			strcpy( buf, "!<could not get current working dir>!");
		std::string tmp(
			"NOTE: Files for dumping are always opened/created in "
			"the \"log\"\ndirectory, please make sure it exsists "
			"and is writable.\nCurrent working directory is \"" +
			std::string( buf) + "\"\n");
		int err = ge.m_thing_manager->dump_all_things( 0,
			(args[1]).c_str());
		switch( err) {
		case 0:
			ge.rcout( "Things dumped to \"" + args[1] + "\"\n");
			return 0;

		case TM_ERR_OPEN_FILE:
			ge.rcout( "Could not open file \"" + args[1] +
				"\" for writing.\n" + tmp);
			return -1;

		case TM_ERR_WRITE_FILE:
			ge.rcout( "Error while writing to file \"" + args[1] +
				"\".\n" + tmp);
			return -2;

		default:
			ge.rcout( "Error " + gm_to_string( err) +
				" while opening/writing file \"" + args[1] +
				"\"\n" + tmp);
			return -3;
		}
	}
	std::string tmp;
	ge.m_thing_manager->dump_all_things( &tmp, 0);
	ge.rcout( tmp);
	return 0;
}

int ccmd_list_spawnable( TMP_ARGS)
{
	ge.rcout( ge.m_thing_manager->list_spawnable());
	return 0;
}

int ccmd_spawn( TMP_ARGS)
{
	if( args.size() != 3) {
		ge.rcout( "usage: spawn <type> <key>\n"
			"\"type\" is the kind of \"thing\" to create/spawn\n"
			"\"key\" is a string that uniquely identifies the \""
				"thing\"\n");
		return -1;
	}
	const int err = ge.m_thing_manager->spawn( args[2], args[1]);
	if( err) {
		ge.rcout( "Error " + gm_to_string( err) +
			" while spawning thing \"" + args[1] +
			"\" with key = \"" + args[2] + "\"\n");
		return -2;
	}
	ge.rcout( "Spawned thing \"" + args[1] + "\" with key = \"" +
		args[2] + "\"\n");
	return 0;
}

int ccmd_remove_thing( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: remove_thing <key>\n"
			"NOTE: Thing \"key\" is deleted and its memory slot "
			"becomes available\nfor other \"things\" to use.\n");
		return -1;
	}
	if( ge.m_thing_manager->remove_thing( args[1])) {
		ge.rcout( "Could not remove thing \"" + args[1] + "\"\n");
		return -2;
	}
	ge.rcout( "Thing \"" + args[1] + "\" removed\n");
	return 0;
}

int ccmd_thing_list_check( TMP_ARGS)
{
	const int err = ge.m_thing_manager->thing_list_check();
	if( err) {
		ge.rcout( "Thing list corrupted. Error = " +
			gm_to_string( err) + "\n");
		return -1;
	}
	ge.rcout( "Thing list is ok\n");
	return 0;
}

int ccmd_use_console_server( TMP_ARGS)
{
	if( !hlp_good_bool_cmd( "use_console_server", args, &ge)) return -1;
	if( ge.m_engine_started) {
		ge.rcout( "\"" + args[0] + "\" only valid in bootup file\n");
		return -2;
	}
	gm_set_state( ge.m_use_console_server, args[1]);
	return 0;
}

int ccmd_cullface( TMP_ARGS)
{
	if( !hlp_good_bool_cmd( "cullface", args, &ge)) return -1;
	if( gm_set_state( args[1])) glEnable( GL_CULL_FACE);
	else glDisable( GL_CULL_FACE);
	return 0;
}

int ccmd_depth_test( TMP_ARGS)
{
	if( !hlp_good_bool_cmd( "depth_test", args, &ge)) return -1;
	if( gm_set_state( args[1])) glEnable( GL_DEPTH_TEST);
	else glDisable( GL_DEPTH_TEST);
	return 0;
}

int ccmd_color_material( TMP_ARGS)
{
	if( !hlp_good_bool_cmd( "color_material", args, &ge)) return -1;
	if( gm_set_state( args[1])) glEnable( GL_COLOR_MATERIAL);
	else glDisable( GL_COLOR_MATERIAL);
	return 0;
}

int ccmd_wireframe( TMP_ARGS)
{
	if( !hlp_good_bool_cmd( "wireframe", args, &ge)) return -1;
	GLint front_back[2];
	glGetIntegerv( GL_POLYGON_MODE, front_back);
	bool b = false;
	if( GL_LINE == front_back[0]) b = true;
	if( gm_set_state( b, args[1]))
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
	else glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
	return 0;
}

int ccmd_perspective( TMP_ARGS)
{
	if( args.size() != 5) {
		ge.rcout( "usage: perspective <feild of view> <aspect ratio> "
			"<near plane> <far plane>\n"
			"NOTE: \"feild of view\" is in degrees.\n"
			"\"near plane\" and \"far plane\" must be positive.\n"
			);
		return -1;
	}
	GLint mat_mode;
	glGetIntegerv( GL_MATRIX_MODE, &mat_mode);
	glMatrixMode( GL_PROJECTION);
	glLoadIdentity();
        gluPerspective(	strtod( (args[1]).c_str(), 0),
			strtod( (args[2]).c_str(), 0),
			strtod( (args[3]).c_str(), 0),
			strtod( (args[4]).c_str(), 0));
	glMatrixMode( mat_mode);
	return 0;
}

int ccmd_gfx_engine( TMP_ARGS)
{
	if( !hlp_good_bool_cmd( "gfx_engine", args, &ge)) return -1;
	gm_set_state( ge.gfx_engine->m_enabled, args[1]);
	return 0;
}

int ccmd_draw_sectors( TMP_ARGS)
{
	if( (args.size() != 2) && (args.size() != 3) &&
		(args.size() != 4)) {
		ge.rcout( "usage: draw_sectors <1/0/on/off> [level] "
				"[line width]\n"
			"NOTE: \"level\" is optional and specifies the "
			"maximum quad tree recursion\ndepth that will be "
				"drawn if sector map drawing is enabled.\n"
			"A draw level of 0, will only draw the first four "
				"quadrants.\n"
			"A higher level will draw more.\n"
			"\"line_width\" is an optional integer describing the "
				"width of rendered sector lines.\n");
		return -1;
	}
	gm_set_state( ge.gfx_engine->m_draw_sectors, args[1]);
	if( 3 == args.size())
		ge.gfx_engine->m_draw_sector_level = atoi( (args[2]).c_str());
	if( 4 == args.size()) {
		ge.gfx_engine->m_sector_line_width = atoi( (args[3]).c_str());
		if( ge.gfx_engine->m_sector_line_width < 1)
			ge.gfx_engine->m_sector_line_width = 1;
	}
	return 0;
}

int ccmd_game_clock( TMP_ARGS)
{
	char buf[64];
	sprintf( buf, "game_clock = %lld\n", ge.game_clock());
	ge.rcout( buf);
	return 0;
}

int ccmd_thing_manager( TMP_ARGS)
{
	std::string tmp;
	ge.m_thing_manager->dbg_stats( tmp);
	ge.rcout( tmp.c_str());
	return 0;
}

int ccmd_list_commands( TMP_ARGS)
{
	if( args.size() > 1) {
		for( int i = 1; i < int( args.size()); i++) {
			ThingControlBlock *tcb = ge.m_thing_manager->get(
				args[i]);
			if( !tcb) {
				ge.rcout( "No such thing \"" + args[i] +
					"\"\n");
			} else {
				ge.rcout( "Commands for \"" + args[i] + "\":\n"
					+ tcb->m_thing->list_ccmd( "\t"));
			}
		}
		return 0;
	}

	std::map <      const std::string,
		int (*)( const std::vector <std::string> &, GameEngine &)>
		:: const_iterator iter;
	iter = ge.m_commands.begin();
	std::string tmp( "Available commands:\n");
	for( ; iter != ge.m_commands.end(); iter++)
		tmp += "\t" + iter->first + "\n";
	tmp += "total commands: " + std::string( gm_to_string( int(
		ge.m_commands.size()))) + "\n";
	ge.rcout( tmp);
	return 0;
}

int ccmd_compile_model_alpha_faces( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: compile_model_alpha_faces <key>\n"
			"NOTE: \"key\" is the name of the model.\n");
		return -1;
	}
	ModelHandle *mh = (ge.gfx_engine->model_man())->get( args[1]);
	if( !mh) {
		ge.rcout( "No such model \"" + args[1] + "\"\n");
		return -2;
	}
	if( !(mh->model)) {
		ge.rcout( "Model \"" + args[1] + "\" is not ready ("
			"model == NULL)\n");
		return -3;
	}
	(mh->model)->compile_alpha_faces();
	ge.rcout( "Alpha faces compiled for model \"" + args[1] + "\"\n");
	return 0;
}

int ccmd_list_models( TMP_ARGS)
{
	ModelManager *mman = ge.gfx_engine->model_man();
	std::map <const std::string, ModelHandle *> :: const_iterator iter =
		mman->get_begin_iterator();
	std::map <const std::string, ModelHandle *> :: const_iterator end =
		mman->get_end_iterator();
	std::string tmp( "Models (reference count, model pointer, key, "
		"file name):\n");
	for( ; iter != end; iter++) {
		tmp += "\t" + gm_to_string(
				(iter->second)->reference_count()) + ", " +
			gm_to_string( (void *)((iter->second)->model)) +
			", \"" + (iter->first) + "\", \"" +
			std::string( (iter->second)->fname) + "\"\n";
	}
	tmp += "Total models: " + gm_to_string( mman->model_count()) + "\n";
	ge.rcout( tmp);
	return 0;
}

int ccmd_dump_model( TMP_ARGS)
{
	if( args.size() < 2) {
		ge.rcout( "usage: dump_model <key> [what to dump] "
			"[what to dump] [what to dump] ...\n"
			"NOTE: If no \"what to dump\" option is given, then "
			"everything is \"dumped\".\n"
			"Things you can dump are \"bones\", \"frames\", "
			"\"textures\", and \"groups\".\n"
			"Also, the numbers of bones/frames/etc. given is "
			"the number of uniquely named (duplicates\n"
			"are not counted nor listed).\n"
			"The index numbers for each element is listed first "
			"followed by the key name.\n");
		return -1;
	}

	ModelHandle *mh = (ge.gfx_engine->model_man())->get( args[1]);
	if( !mh) {
		ge.rcout( "No such model with key = \"" + args[1] + "\"\n");
		return -2;
	}

	if( args.size() > 2) {
		for( int i = 2; i < int( args.size()); i++) {
			if( "bones" == args[i])
				ge.rcout( mh->dump_bones());

			else if( "frames" == args[i])
				ge.rcout( mh->dump_frames());

			else if( "textures" == args[i])
				ge.rcout( mh->dump_textures());

			else if( "groups" == args[i])
				ge.rcout( mh->dump_groups());

			else ge.rcout( "Can't dump \"" + args[i] + "\"\n");
		}

	} else ge.rcout( mh->dump());

	return 0;
}

int ccmd_remove_model( TMP_ARGS)
{
	if( 2 != args.size()) {
		ge.rcout( "usage: remove_model <key>\n"
			"NOTE: Models are only removed if they are not being "
			"referenced by\nany \"thing\".\n"
			"Also, type \"list_models\" with no parameters to see "
			"which models\nare available for removal.\n");
		return -1;
	}
	int err = (ge.gfx_engine->model_man())->remove( args[1]);
	switch( err) {
	case 0:
		ge.rcout( "Model with key = \"" + args[1] +
			"\" has been removed.\n");
		break;

	case MODEL_MANAGER_ERR_NO_MODEL:
		ge.rcout( "There is no model with key = \"" + args[1] +
			"\"\n");
		return -1;

	case MODEL_MANAGER_ERR_REFERENCE:
		ge.rcout( "Model with key = \"" + args[1] + "\" still has " +
			gm_to_string(
				((ge.gfx_engine->model_man())->get( args[1]))->
					reference_count()) +
			" reference(s),\nmodel will not be removed.\n");
		return -2;

	default:
		ge.rcout( "Error " + gm_to_string( err) +
			" while removing model with key = \"" + args[1] +
			"\"\n");
		return -3;
	}
	return 0;
}

int ccmd_load_model_textures( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: load_model_textures <key>\n");
		return -1;
	}
	ModelHandle *h = (ge.gfx_engine->model_man())->get( args[1]);
	if( !h) {
		ge.rcout( "There is no model with key = \"" +
			args[1] + "\"\n");
		return -2;
	}
	const char *paths[] = { "./media", "./media/bmp"};
	if( h->model->load_textures( paths, 2)) {
		ge.rcout( "Not all textures for model with key = \"" +
			args[1] + "\" could be loaded\n");
		return -3;
	}
	ge.rcout( "textures reloaded for model with key = \"" +
		args[1] + "\"\n");
	return 0;
}

int ccmd_set_alt_texture( TMP_ARGS)
{
	if( args.size() != 4) {
		ge.rcout( "usage: set_alt_texture <model> <des> <src>\n"
			"NOTE: \"model\" is the model that will receive the "
				"alternate texture.\n"
			"\"des\" is the which texture from the model that "
				"will receive the alternate texture.\n"
			"\"src\" is the name of a pre-loaded texture that "
				"will be the model's alternate texture.\n");
		return -1;
	}
	ModelHandle *mdl = (ge.gfx_engine->model_man())->get( args[1]);
	if( !mdl) {
		ge.rcout( "No such model \"" + args[1] + "\"\n");
		return -2;
	}
	const int err = mdl->set_alternate_texture( args[2], args[3],
		ge.gfx_engine);
	if( err) {
		ge.rcout( "Error " + gm_to_string( err) + " while setting "
			"alternate texture \"" + args[3] + "\" for model \"" +
			args[1] + "\" (des = \"" + args[2] + "\").\n");
		return -3;
	}
	return 0;
}

int ccmd_set_texture_color_key( TMP_ARGS)
{
	if( args.size() != 5) {
		ge.rcout( "usage: set_texture_color_key <key> <red> <green> "
			"<blue>\n"
			"NOTE: \"key\" is the unique name of the texture.\n"
			"Color components red, green and blue are integers "
				"in the range [0, 255].\n"
			"example: set_texture_color_key my_texture 50 65 30\n"
			);
		return -1;
	}
	TextureHandle *t = ge.gfx_engine->get_texture( args[1]);
	if( !t) {
		ge.rcout( "No such texture \"" + args[1] + "\"\n");
		return -2;
	}
	if( !(t->m_tex)) {
		ge.rcout( "Texture \"" + args[1] + "\" not ready ("
			"m_tex == NULL)\n");
		return -3;
	}
	t->m_tex->setColorKey(	atoi( (args[2]).c_str()),
				atoi( (args[3]).c_str()),
				atoi( (args[4]).c_str()));
	return 0;
}

int ccmd_load_texture( TMP_ARGS)
{
	if( args.size() < 2) {
		ge.rcout( "usage: load_texture <key> [file name] "
				"[load options]\n"
			"NOTE: Texture with key \"key\", will be created "
				"(if not created already)\n"
			" and (optionaly) will load the texture \"file name\""
				"\n"
			"Giving only the \"key\" argument will create a "
				"blank texture.\n");
		return -1;
	}
	TextureHandle *t = ge.gfx_engine->add_texture( args[1]);
	if( !t) {
		ge.rcout( "Could not create texture with key = \"" + args[1]
			+ "\"\n");
		return -2;
	}
	if( 2 == args.size()) return 0;

	if( !(t->m_tex)) {
		ge.rcout( "Texture not ready. \"m_tex\" == NULL.\n");
		return -3;
	}
	// Check texture-load options arguments.
	int flags = RGGL_LARGE_TEXTURE;
	for( int i = 3; i < int( args.size()); i++) {
		if( "smooth" == args[i]) flags |= RGGL_SMOOTH_TEXTURE;
		else if( "mipmap" == args[i]) flags |= RGGL_MIPMAP;
		else if( "colorkey" == args[i]) flags |= RGGL_COLOR_KEY;
		else {
			ge.rcout( "Invalid texture option \"" + args[i] +
				"\"\n");
		}
	}
	const int err = t->m_tex->load( (args[2]).c_str(), flags);
	if( err) {
		ge.rcout( "Could not load texture \"" + args[2] + "\"\n"
			"\tError: \"" + rgErrorString( err) + "\"\n");
		return -4;
	}
	ge.rcout( "Loaded texture \"" + args[2] + "\"\n");
	return 0;
}

int ccmd_load_model( TMP_ARGS)
{
	ModelHandle *h;
	if( 2 == args.size()) {
		h = (ge.gfx_engine->model_man())->get( args[1]);
		if( !h) {
			ge.rcout( "There is no model with key = \"" +
				args[1] + "\"\n");
			return -4;
		}

	} else if( args.size() != 3) {
		ge.rcout( "usage 1: load_model <key>\n"
			"usage 2: load_model <key> <filename>\n"
			"NOTE: The first usage will reload an existing model."
			"\nThe second usage will load a new model.\n"
			"\"key\" is a string which identifies the model,"
			" it must be unique.\n"
			"\"filename\" can also include a path.\n"
			"example 1: \"load_model md1 my_model.3dr\"\n"
			"example 2: \"load_model md1 ./media/my_model.3dr\"\n"
			);
		return -1;

	} else {
		ModelManager *m = (ge.gfx_engine->model_man());
		h = m->get( args[1]);
		if( h) {
			ge.rcout( "Model with file name \"" +
				std::string( h->fname) +
				"\" already has key = \"" + args[1] + "\"\n");
			return -2;
		}
		h = m->add( args[1], args[2]);
		if( !h) {
			ge.rcout( "Could not add/create space for model \"" +
				args[2] +
				"\" with key = \"" + args[1] + "\"\n");
			return -3;
		}
	}
	const int err = h->load();
	if( err) {
		ge.rcout( "Error " + gm_to_string( err) +
			" while loading model \"" + args[2] +
			"\" with key = \"" + args[1] + "\"\n");
		return -4;
	}
	ge.rcout( "Loaded model \"" + args[2] + "\"\n");
	return 0;
}

int ccmd_set_root_renderer( TMP_ARGS)
{
	if( args.size() != 2) {
		ge.rcout( "usage: root_renderer <key>\n"
			"NOTE: \"key\" is the thing to make the root "
			"renderer.\n");
		return -1;
	}
	if( !(ge.gfx_engine->set_root_renderer( args[1]))) {
		ge.rcout( "Could not set thing with key = \"" + args[1] +
			"\" to be the root_renderer\n");
		return -2;
	}
	ge.rcout( "thing \"" + args[1] + "\" is now the root renderer\n");
	return 0;
}

int ccmd_load_module( TMP_ARGS)
{
	if( args.size() != 3) {
		ge.rcout( "usage: load_module <key> <filename>\n"
			"NOTE: \"key\" uniquely identifies the module.\n"
			"\"filename\" is the module to load.\n");
		return -1;
	}
	std::string emsg;
	const int err = ge.m_thing_manager->load_module( args[1], args[2],
		emsg);
	if( err) {
		ge.rcout( "Error " + gm_to_string( err) +
			" while loading module \"" + args[2] + "\"\n"
			"Error message: \"" + emsg + "\"\n");
		return -2;
	}
	ge.rcout( "Module \"" + args[2] + "\", key = \"" + args[1] +
		"\" loaded\n");
	return 0;
}

#undef TMP_ARGS
