/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "things.h"
#include "engine/engine_game.h"
#include "engine/engine_physics.h"
#include "engine/engine_graphics.h"
#include <fstream>
#include <rg3DRom.h>
#include <rg3DRom_csg.h>
#include <rg3DRom_physics.h>

std::map <const std::string, int (*)( const gmCCArg &)> gmPhyModel::m_ccmd;

extern "C" gmThing *spw_physical_model()
{
        return new gmPhyModel();
}

extern "C" int del_physical_model( gmThing *t)
{
	delete t;
	return 0;
}

gmPhyModel::gmPhyModel()
{
	m_use_gravity = true;
	m_skel = 0;

	m_cc = new PhyCollisionContext( this);

	m_bounciness = 0.5f;

	m_draw_dyn_bboxes = false;

	if( !(m_ccmd.empty())) return;
        add_console_command( "set_model", gmPhyModel_set_model, m_ccmd);
	add_console_command( "load_identity", gmPhyModel_load_identity,
		m_ccmd);
	add_console_command( "scale", gmPhyModel_scale, m_ccmd);
        add_console_command( "set_center_of_mass",
		gmPhyModel_set_center_of_mass, m_ccmd);
        add_console_command( "add_dynamic_bboxes",
		gmPhyModel_add_dynamic_bboxes, m_ccmd);
	add_console_command( "dump_dynamic_bboxes",
		gmPhyModel_dump_dynamic_bboxes, m_ccmd);
	add_console_command( "draw_dynamic_bboxes",
		gmPhyModel_draw_dynamic_bboxes, m_ccmd);
	add_console_command( "bounciness", gmPhyModel_bounciness, m_ccmd);
}

gmPhyModel::~gmPhyModel()
{
	if( m_cc) delete m_cc;
	if( m_skel) delete m_skel;
}

void gmPhyModel::process( ThingManager *tm)
{
	// Get models physical state before transformations.
	// Also, save current model skeleton since applying transformations may
	// put the skeleton in an invalid (colliding) physical state. The
	// skeleton "m_draw_context->m_skout" is saved along with supporting
	// matrices.

	PhysicsEngine *pe = phy_engine();
	pe->get_origin_points( *m_cc);
	m_draw_context->backup_main_skel();


	// Apply transformations to the model.
	animate();
	apply_gravity();
	apply_velocity();


	// Get a physics context (so physical stuff can be applied to the
	// model). Also, get the models physical state after the applied
	// transformations.
	if( model()) (model())->skel.update_fcs( *m_draw_context,
		m_draw_context->m_skout);
	pe->get_bounding_points( *m_cc);


	// Check if the transformed skeleton is in an invalid phyiscal state.
	// If it is, then perform "anti-intrusion" functions on it if possible.
	// The "thing" gets "attempts" number of tries to produce a physically
	// valid transformation. If all attempts fail, then the original
	// skeleton is restored (since it IS in a valid physical state).
	int attempts = 4;
	PhyDynResult result;
	const rgVector velocity_backup( m_velocity);
	while(	pe->environment_collision( result, *m_cc) &&
		((attempts--) > 0)) {

		// Adjust velocity based on collision.
		rgVector v;
		rgVector avg_surface_normal;
		rgVector anti_intrude;
		m_velocity = velocity_backup;
		for( int i = 0; i < int( result.m_point_hit.size()); i++) {
			PhyCollisionFragment &f = result.m_point_hit[i];
			//v += post_collision_vector( f.intrusion_vector,
			//	f.surface_normal);
			avg_surface_normal += f.surface_normal;
			anti_intrude += anti_intrusion( f);
			const float k = m_velocity.dot( f.surface_normal);
			if( k < 0) {
				const rgVector &n = f.surface_normal;
				rgVector &v = m_velocity;
				v -= (k * (1.0f + m_bounciness)) * n;

				// FIXME: Do proper friction
				const float len = RG_MIN(
					f.prisim->m_friction,
					v.length());
				if( len < 0.00001f) v = rgVector();
				else {
					v -= ((v - (v.dot( n) * n)).
						normalize()) *
						(len /
					float( (ge())->game_clock_hz()));
				}
//					(len /
//					pe->friction_frame_scalar());
//				}

				//const rgVector bounce(
				//	(k - m_bounciness) * f.surface_normal);
				//std::cerr << "vel = " << m_velocity << "\n";
				//std::cerr << "surf = " << f.surface_normal
				//	<< "\n";
				//std::cerr << "k = " << k << "\n";
				//m_velocity += (f.su * k) - bounce;
				//m_velocity -= bounce;
				//std::cerr << "vel = " << m_velocity << "\n";
				//std::cerr << "bounce = " << bounce << "\n";
				//exit(0);
			}
		}
		const float k = 1.0f / float( result.m_point_hit.size());
		anti_intrude = anti_intrude * k;
		v = v * k;
		avg_surface_normal.normalize();

		if( attempts > 0) {
			m_draw_context->translate( anti_intrude);
			//m_draw_context->translate( anti_intrude +
			//	(rgVector( anti_intrude).normalize() * 0.001));
			//std::cerr << "anti_intrude = " <<
			//	anti_intrude << "\n";
			//apply_velocity();

			//m_draw_context->update_matrix();
			if( model()) (model())->skel.update_fcs(
				*m_draw_context, m_draw_context->m_skout);
			pe->get_bounding_points( *m_cc);
		}
	}
	if( attempts < 0) {
		m_draw_context->restore_main_skel();
		m_velocity = rgVector();
	}
}

rgVector gmPhyModel::anti_intrusion( PhyCollisionFragment &f)
{
	const rgVector aip( f.collision_point + (f.surface_normal *
		(f.radius + 0.001f)));
	return (f.surface_normal * f.surface_normal.dot( (aip -
		f.des_point)));
	// Apply friction
//	const rgVector fric( (aip - pos) * f.prisim->m_friction);
//	return pos + fric;
}

rgVector gmPhyModel::post_collision_vector( const rgVector v, const rgVector n)
	const
{
	const rgVector r( rgReflectAbout( -v, n));
	const rgVector p( (v + r) * 0.5f);
	return ((r * m_bounciness) + (p * (1.0f - m_bounciness)));
/*
	const rgVector q( ((r * m_bounciness) + (p * (1.0f - m_bounciness))));
	std::cout << "----------------------------\n";
	std::cout << "v = " << v << "\n";
	std::cout << "n = " << n << "\n";
	std::cout << "r = " << r << "\n";
	std::cout << "p = " << p << "\n";
	std::cout << "q = " << q << "\n";
	exit(0);
	return q;
	*/
}

void gmPhyModel::draw( gmRenderArgs *ra)
{
	gmModelThing::draw( ra);
	if( m_draw_dyn_bboxes) draw_dynamic_bboxes();
}

void gmPhyModel::draw_dynamic_bboxes()
{
	sk3DRomModel *m = model();
	if( !m) return;
	if( !(m->m_csg)) return;

	// No blend
	GLboolean blend;
	glGetBooleanv( GL_BLEND, &blend);
	glDisable( GL_BLEND);

	// No texturing
	GLboolean texture;
	glGetBooleanv( GL_TEXTURE_2D, &texture);
	glDisable( GL_TEXTURE_2D);

	glBegin( GL_QUADS);
	glColor3f( 0, 0, 1);
	for( int i = 0; i < int( m_dyn_bbox.size()); i++) {
		std::vector <rgVector> v;
		get_bbox( v, m_dyn_bbox[i]);
		rgCollisionPrisim tmp;
		tmp.m_v[0] = v[0];
		tmp.m_v[1] = v[1];
		tmp.m_v[2] = v[2];
		tmp.m_v[3] = v[3];
		tmp.m_v[4] = v[4];
		tmp.m_v[5] = v[5];
		tmp.m_v[6] = v[6];
		tmp.m_v[7] = v[7];
		tmp.render();
	}
	glEnd();

	if( blend) glEnable( GL_BLEND);
	if( texture) glEnable( GL_TEXTURE_2D);
}

int gmPhyModel::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end())
		return gmModelThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string gmPhyModel::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++) {
		tmp += p_prefix + gmPhyModel::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += gmModelThing::list_ccmd( p_prefix);
	return tmp;
}

int gmPhyModel_set_model( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;

	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( gmModelThing_set_model( p)) return GM_THING_ERR_GENERIC;
	if( !(pm->model())) return GM_THING_ERR_GENERIC;
	if( pm->m_skel) { delete pm->m_skel; pm->m_skel = 0;}
	pm->m_skel = new sk3DRomPhySkel( (pm->model())->skel,
		*(pm->m_draw_context));
	return 0;
}

// gmPhyModel_load_identity is required to override gmModelThing_load_identity
// because loading the identity matrix may affect the sizes of the bounding
// boxes.
int gmPhyModel_load_identity( const gmCCArg &p)
{
	if( gmModelThing_load_identity( p)) return GM_THING_ERR_GENERIC;
	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	(pm->phy_engine())->gen_bounding_points( *(pm->m_cc));
	return 0;
}

// gmPhyModel_scale is required to override gmModelThing_scale because scaling
// the model affects the sizes of the bounding boxes.
int gmPhyModel_scale( const gmCCArg &p)
{
	if( gmModelThing_scale( p)) return GM_THING_ERR_GENERIC;
	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	(pm->phy_engine())->gen_bounding_points( *(pm->m_cc));
	return 0;
}

int gmPhyModel_bounciness( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( 0 == args.size()) {
		p.ge->rcout( "Bounciness = " + gm_to_string( pm->m_bounciness)
			+ "\n");
	} else pm->m_bounciness = strtod( (args[0]).c_str(), 0);
	return 0;
}

int gmPhyModel_set_center_of_mass( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;

	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( !pm->m_mh) return GM_THING_ERR_GENERIC;
	const int bid = (pm->m_com.m_bone_id = pm->m_mh->bone_id( args[0]));
	if( bid < 0) {
		p.ge->rcout( "No bone named \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	if( !(pm->model())) return GM_THING_ERR_GENERIC;
	pm->m_com.m_bound_position =
		((pm->model())->skel.b[bid]).get_jcs_pivot();
	return 0;
}

int gmPhyModel_add_dynamic_bboxes( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: add_dynmaic_bboxes <group key>\n");
		return GM_THING_ERR_PARAM;
	}

	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( !(pm->add_dynamic_bboxes( args[0]))) {
		p.ge->rcout( "gmPhyModel_add_dynamic_bboxes:\n"
			"Could not add bboxes with group = \"" + args[0] +
			"\"\n");
		return GM_THING_ERR_GENERIC;
	}
	return 0;
}

int gmPhyModel_dump_dynamic_bboxes( const gmCCArg &p)
{
	p.ge->rcout( ((gmPhyModel *)(p.thing))->dump_dyn_bboxes());
	return 0;
}

int gmPhyModel_draw_dynamic_bboxes( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: draw_dynmaic_bboxes <0/1/on/off/toggle>\n"
			);
		return GM_THING_ERR_PARAM;
	}
	bool &b = ((gmPhyModel *)(p.thing))->m_draw_dyn_bboxes;
	b = gm_set_state( b, args[0]);
	return 0;
}

// Description: Returns a vector of absolute vertex locations representing the
//	bounding box. The model must be transformed at least once by the
//	"thing" inorder to get defined results for the bounding box.
//
// FIXME: The below comment, which deals with the "fcs" (Full Coordinate
//	System) issue has been solved. sk3DRomModel now accepts draw contexts
//	(sk3DRomDrawContext) which allow a "thing" to keep its own copy of the
//	bone "fcs"'s. This is done transparently when ever a "thing" provides
//	its draw context to the model when transforming it. So get_bbox now
//	just uses the "fcs" stored in the draw context rather than the model.
//
//	The vertices cannot be calculated until the 
//	"sk3DRomMode::skel.b[n].fcs" has been calculated for all bones
//	containing bounding boxes. "fcs" stands for (Full Coordinate System).
//	This means the model has to transformed before bounding boxes can
//	be retrieved. This is not convinient (since it would be better to be
//	able to access bounding boxes at random). Also, this forces "things"
//	to retransform their models at every frame (since "fcs" will be
//	corrupted by other "things" using the same model).
//
//	This problem is solved by forcing the "thing" to save the "fcs" of all
//	required bones and then restoring the "fcs" for those bones before
//	it uses the skeleton again.
int gmPhyModel::get_bbox( std::vector <rgVector> &v, int id)
{
	v.clear();
	sk3DRomModel *m = model();
	if( !m) return -1;
	if( !(m->m_csg)) return -2;
	const rgCSGBrush *b = m->m_csg->brush( id);
	if( !b) return -3;
	// FIXME: "b" should ALWAYS be pointing to a "RG3DROM_CSG_CUBE", so
	//	the "if" check shouldn't be neccasary.
	if( RG3DROM_CSG_CUBE == b->type()) {
		b->map_verts( v, m_draw_context->get_fcs(
			((sk3DRomCSGCube *)b)->m_bone_id).map_csys(
				b->raw_coord_system()));

	} else b->gen_verts( v);
	const rgMatrix &mat = m_draw_context->mat();
	for( int i = 0; i < int( v.size()); i++)
		v[i] = mat.pointMul( v[i]);
	return 0;
}

rgVector gmPhyModel::get_bbox_center( const int id)
{
	if( !(model())) return rgVector();
	const rgCSGBrush *b = (model())->m_csg->brush( id);
	// FIXME: "b" should ALWAYS be pointing to a "RG3DROM_CSG_CUBE", so the
	//	"if" check shouldn't be neccasary.
	if( RG3DROM_CSG_CUBE == b->type()) {
		return (m_draw_context->mat()).pointMul(
			((m_draw_context->get_fcs(
				((sk3DRomCSGCube *)b)->m_bone_id)).map_csys(
					b->raw_coord_system())).p);
	}
	return rgVector();
}

PhyBoundingPoint gmPhyModel::get_bbox_point( const int id)
{
	if( !(model())) return PhyBoundingPoint();
	const rgCSGBrush *b = (model())->m_csg->brush( id);
	// FIXME: "b" should ALWAYS be pointing to a "RG3DROM_CSG_CUBE", so the
	//	"if" check shouldn't be neccasary.
	if( RG3DROM_CSG_CUBE == b->type()) {
		const rgMatrix mat( m_draw_context->mat());
		return PhyBoundingPoint(
				mat.pointMul(
				((m_draw_context->get_fcs(
				((sk3DRomCSGCube *)b)->m_bone_id)).map_csys(
					b->raw_coord_system())).p),

				// XXX: Bounding prisim's width is not affected
				//	by bone transformations.
				// FIXME: Calculating the length every time a
				//	point is accessed is not efficient.
				//	Use precalculation or use a more
				//	efficient method.
				(mat * (b->raw_coord_system()).side).length()
			);
	}
	return PhyBoundingPoint();
}

std::string gmPhyModel::dump_dyn_bboxes()
{
	std::string tmp( "Dynamic Bounding Boxes (Cube Brushes): " +
		gm_to_string( int( m_dyn_bbox.size())) + "\n");
	sk3DRomModel *m = model();
	if( !m) return tmp;
	if( !(m->m_csg)) return tmp;
	for( int i = 0; i < int( m_dyn_bbox.size()); i++) {
		std::vector <rgVector> v;
		if( get_bbox( v, m_dyn_bbox[i])) {
			tmp += "\tCould not get brush with id = " +
				gm_to_string( m_dyn_bbox[i]) + "\n";

		} else {
			if( v.empty()) {
				tmp += "\tVertex-less brush, id = " +
					gm_to_string( m_dyn_bbox[i]) + "\n";
			} else {
				tmp += "\tBrush id = " + gm_to_string(
						m_dyn_bbox[i]) + ", " +
					gm_to_string( int( v.size())) +
					" vertices\n";
				while( !(v.empty())) {
					tmp += "\t\t" + gm_to_string( v[0]) +
						"\n";
					v.erase( v.begin());
				}
			}
		}
	}
	return tmp;
}

gmPhyModelCOM &gmPhyModel::center_of_mass()
{
	if( !m_skel) return m_com;
	const sk3DRomSkel *s = m_skel->skel();
	if( m_draw_context) {
		m_com.m_position =
			(m_draw_context->mat()).pointMul(
				(s->get_full_bone_rcs( m_com.m_bone_id,
				m_draw_context->m_skout)).map_point(
					m_com.m_bound_position));

	}
	return m_com;
}


// Description: Apply gravity to the bone that contains the center-of-mass
//	(COM). That bone, will then take all other bones with it (thus
//	creating the effect that gravity is applied to all bones).
void gmPhyModel::apply_gravity()
{
	if( !m_use_gravity) return;
	m_velocity += (phy_engine())->m_gravity;
}

void gmPhyModel::apply_velocity()
{
	//if( !m_skel) return;
	//const sk3DRomSkel *s = m_skel->skel();
	//s->translate_rcs( m_com.m_bone_id, m_draw_context->m_skout,
	//	m_velocity);
	m_draw_context->translate( m_velocity);
//	m_draw_context->update_matrix();
}

bool gmPhyModel::add_dynamic_bboxes( const std::string &key)
{
	if( !m_mh) return false;
	const int gid = m_mh->group_id( key);
	if( gid < 0) return false;
	sk3DRomModel *m = model();
	if( !m) return false;
	if( !(draw_context())) return false;
	for( int i = 0; i < m->m_csg->brush_count(); i++) {
		rgCSGBrush *b = m->m_csg->brush( i);
		if(	(gid == b->group_id()) &&
			(RG3DROM_CSG_CUBE == b->type())) {
			m_dyn_bbox.push_back( i);
		}
	}
	return true;
}
                                                                                
void gmPhyModel::clear_dynamic_bboxes()
{
	m_dyn_bbox.clear();
}
