/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "things.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
#include <fstream>

#ifdef UNIX_LIKE_OS
#include <dlfcn.h>
#endif

void phDebugger::process( ThingManager *tm)
{
	std::cerr << "process debugger\n";
}

ThingManager::ThingManager( GameEngine *p_game_engine)
{
	m_game_engine = p_game_engine;

	m_max_things = TM_MAX_THINGS;
	m_things = new ThingControlBlock[ m_max_things];
        init_things();

	init_spawn_list();
}

ThingManager::~ThingManager()
{
	clear();
}

void ThingManager::clear()
{
	while( m_thing_head >= 0) {
		const int err = remove_thing( (m_things[ m_thing_head]).m_key);
		if( err) {
			std::cerr << "ThingManager::clear(): Error " << err
				<< " while trying to delete thing \"" <<
				(m_things[ m_thing_head]).m_key <<
				"\"\n\taborting (not deleting remaining \""
				"things\")\n\tThing Dump:\n" <<
				dump_thing( (m_things[ m_thing_head]).m_key);
			break;
		}
	}
}


int ThingManager::create_thing( const std::string &key, gmThing *thg,
	int p_pid)
{
	if( get( key)) return TM_ERR_SAME_KEY;
	if( !thg) return TM_ERR_GENERIC;
	if( m_thing_free < 0) return -2;
	const int tcb_index = m_thing_free;
	typedef std::pair <const std::string, int> SIPair;
	if( false == (m_thing_map.insert( SIPair( key, tcb_index)).second))
		return TM_ERR_GENERIC;
	ThingControlBlock &tcb = m_things[ m_thing_free];
	m_thing_count++;

	// Set the new head free block.
	m_thing_free = tcb.m_next;
	if( m_thing_free >= 0) {
		ThingControlBlock &fcb = m_things[ m_thing_free];
		fcb.m_prev = -1;
	}

	// The new tcb will be the at the head of the "things" list.
	tcb.m_next = m_thing_head;
	(m_things[ m_thing_head]).m_prev = tcb_index;
	m_thing_head = tcb_index;
	tcb.m_prev = -1;
	tcb.m_thing = thg;
	tcb.m_thing->m_tcb = &tcb;
	tcb.m_key = key;
	tcb.m_free = false;
	tcb.m_manager = this;
	move_thing_to_proc_list( &tcb, p_pid);
	return 0;
}

int ThingManager::remove_thing( const std::string &key)
{
	ThingControlBlock *tcb = get( key);
	if( !tcb) return -1;		// No "thing"
	if( tcb->is_free()) return -2;	// "Thing" already removed
	if( !(tcb->m_thing)) return -3;	// No "thing" to remove
	std::map <const std::string, int (*)( gmThing *)> :: const_iterator
		iter = m_delete.find( (tcb->m_thing)->type());
	if( iter == m_delete.end()) return -4;	// No delete function
	const int err = (iter->second)( tcb->m_thing);
	if( err) return -5;
	tcb->m_thing = 0;
	tcb->m_key.clear();
	tcb->m_free = true;
	m_thing_count--;

	// Detach from allocated thing list.
	if( index( tcb) == m_thing_head) {
		if( tcb->m_prev >= 0) m_thing_head = tcb->m_prev;
		else m_thing_head = tcb->m_next;
	}
	if( tcb->m_next >= 0) (m_things[ tcb->m_next]).m_prev = tcb->m_prev;
	if( tcb->m_prev >= 0) (m_things[ tcb->m_prev]).m_next = tcb->m_next;

	// Attach to free list
	const int tmp = m_thing_free;
	m_thing_free = index( tcb);
	(m_things[ m_thing_free]).m_prev = -1;
	(m_things[ m_thing_free]).m_next = tmp;
	if( tmp >= 0) (m_things[ tmp]).m_prev = m_thing_free;
	return 0;
}

int ThingManager::thing_list_check() const
{
	int tmp_alloc_count = 0;
	{ // Check allocated things
		int i = m_thing_head;
		while( i >= 0) {
			tmp_alloc_count++;
			i = (m_things[i]).m_next;
		}
		if( m_thing_count != tmp_alloc_count) return -1;
	}

	int tmp_free_count = 0;
	{ // Check free list
		int i = m_thing_free;
		while( i >= 0) {
			tmp_free_count++;
			i = (m_things[i]).m_next;
		}
		if( (m_max_things - m_thing_count) != tmp_free_count)
			return -2;
	}
	if( m_max_things != (tmp_alloc_count + tmp_free_count)) return -3;
	return 0;
}

void ThingManager::move_thing_to_proc_list( ThingControlBlock *tcb, int pid)
{
	if( (pid < 0) || (pid >= TM_MAX_PROC_LIST)) pid = TM_MAX_PROC_LIST - 1;
	move_thing_to_proc( *tcb, &(m_proc[ pid]));
}

void ThingManager::move_thing_to_proc( ThingControlBlock &tcb,
	std::list <ThingControlBlock *> *p_proc)
{
	if( tcb.m_proc == p_proc) return;
	remove_thing_from_proc( tcb);
	if( !p_proc) return;
	p_proc->push_back( &tcb);
	tcb.m_proc = p_proc;
	(tcb.m_iter = p_proc->end())--;
}

void ThingManager::remove_thing_from_proc( ThingControlBlock &tcb)
{
	if( !(tcb.m_proc)) return;
	tcb.m_proc->erase( tcb.m_iter);
	tcb.m_proc = 0;
}

void ThingManager::process()
{
	process_list( m_proc[0]);
	if( !((ge())->game_clock() & 0x1)) process_list( m_proc[1]);
	if( !((ge())->game_clock() & 0x3)) process_list( m_proc[2]);
	if( !((ge())->game_clock() & 0x7)) process_list( m_proc[3]);
	if( !((ge())->game_clock() & 0xF)) process_list( m_proc[4]);
	if( !((ge())->game_clock() & 0x1F)) process_list( m_proc[5]);
	if( !((ge())->game_clock() & 0x3F)) process_list( m_proc[6]);
	if( !((ge())->game_clock() & 0x7F)) process_list( m_proc[7]);
	if( !((ge())->game_clock() & 0xFF)) process_list( m_proc[8]);
	if( !((ge())->game_clock() & 0x1FF)) process_list( m_proc[9]);
	if( !((ge())->game_clock() & 0x3FF)) process_list( m_proc[10]);
	if( !((ge())->game_clock() & 0x7FF)) process_list( m_proc[11]);
	if( !((ge())->game_clock() & 0xFFF)) process_list( m_proc[12]);
	if( !((ge())->game_clock() & 0x1FFF)) process_list( m_proc[13]);
	if( !((ge())->game_clock() & 0x3FFF)) process_list( m_proc[14]);
	if( !((ge())->game_clock() & 0x7FFF)) process_list( m_proc[15]);
	// XXX: process_list( m_proc[16]) is the sleeping list.
}

void ThingManager::process_list( std::list <ThingControlBlock *> &p_proc)
{
	std::list <ThingControlBlock *>::iterator iter;
	for( iter = p_proc.begin(); iter != p_proc.end(); iter++)
		if( (*iter)->m_thing) (*iter)->m_thing->process( this);
}

void ThingManager::init_things()
{
	m_thing_count = 0;
	m_thing_head = -1;
	m_thing_free = 0;
	for( int i = 0; i < m_max_things; i++) {
		m_things[i].m_prev = i - 1;
		m_things[i].m_next = i + 1;
		m_things[i].m_free = true;
	}
	m_things[ m_max_things - 1].m_next = -1;
}

void ThingManager::dbg_stats( std::string &des) const
{
	des.clear();
	char buf[256];
	sprintf( buf,
		"ThingManager Statistics:\n"
		"current things: %d, max things = %d, free = %d\n"
		"thing head: %d\n"
		"free list head: %d\n",
		count(), m_max_things, m_max_things - count(),
		m_thing_head,
		m_thing_free);
	des = buf;
}

std::string ThingManager::dump_thing( const std::string &key) const
{
	ThingControlBlock *tcb = get( key);
	if( !tcb) {
		return std::string( "no such thing with key = \"" + key +
			"\"\n");
	}
	return dump_thing( tcb - m_things);
}

std::string ThingManager::dump_thing( const int id) const
{
	if( (id < 0) || (id >= m_max_things)) {
		return std::string( "no such thing with id = " +
			gm_to_string( id) + "\n");
	}
	
	ThingControlBlock *tcb = &(m_things[ id]);
	std::string tmp(
		"Thing " + gm_to_string( (void *)tcb) +
			", index = " + gm_to_string( index( tcb)) + ":\n");
	if( !(tcb->m_thing)) tmp += "\ttype = null\n";
	else tmp += "\ttype = \"" + tcb->m_thing->type() + "\"\n";
	tmp +=	"\tkey = \"" + tcb->m_key + "\"\n" +
		"\tprev index = " + gm_to_string( tcb->m_prev) + "\n" +
		"\tnext index = " + gm_to_string( tcb->m_next) + "\n" +
		"\tm_proc = " + gm_to_string( (void *)m_proc) + "\n";
	return tmp + "\n";
}

int ThingManager::dump_all_things( std::string *s, const char *fname) const
{
	if( s) {
		int i = m_thing_head;
		int num = 0;
		(*s) = "Thing Dump: Reported count = " + gm_to_string( count())
			+ "\n";
		while( i >= 0) {
			(*s) += dump_thing( i);
			i = (m_things[i]).m_next;
			num++;
		}
		(*s) += "things dumped: " + gm_to_string( num) + "\n";

	} else if( fname) {
		std::ofstream des( ("./log/" + std::string( fname)).c_str());
		if( !(des.is_open())) return TM_ERR_OPEN_FILE;
		const std::string tmp = rg_get_extension( fname);
		if( ("htm" == tmp) || ("html" == tmp))
			html_dump_all_things( des);

		else {
			des << "Thing Dump: Reported count = " << count() <<
				"\n";
			int num = 0;
			int i = m_thing_head;
			while( i >= 0) {
				des << dump_thing( i);
				i = (m_things[i]).m_next;
				num++;
			}
			des << "things dumped: " << num << "\n";
		}
		if( des.fail()) return TM_ERR_WRITE_FILE;
	}
	return 0;
}

void ThingManager::html_dump_all_things( std::ostream &des) const
{
	des << "<html>\n";
	des << "<head>\n";
	des << "\t<title>\n";
	des << "ThingManager \"thing\" dump\n";
	des << "\t</title>\n";
	des << "</head>\n";

	des << "<body>\n";
	des << "<h3>Total reported things: " << count() << "</h3>\n";
	int num = 0;
	int i = m_thing_head;
	while( i >= 0) {
		des << html_dump_thing( i);
		i = (m_things[i]).m_next;
		num++;
	}
	des << "<h3>Total things dumped: " << num << "</h3>\n";
	des << "</body>\n";

	des << "</html>";
}

std::string ThingManager::html_dump_thing( const int id) const
{
	if( (id < 0) || (id >= m_max_things)) {
		return std::string( "<p>no such thing with id = " +
			gm_to_string( id) + "</p>\n");
	}
	
	ThingControlBlock *tcb = &(m_things[ id]);
	std::string tmp(
		"<hr><p>\n"
		"<a name=\"" + gm_to_string( index( tcb)) + "\"></a>\n" +
		"Thing " + gm_to_string( (void *)tcb) +
			", index = " + gm_to_string( index( tcb)) + ":\n");
	tmp += "<dl>\n";
	if( !(tcb->m_thing)) tmp += "<dd>type = null\n";
	else tmp += "<dd>type = \"" + tcb->m_thing->type() + "\"\n";
	tmp +=	"<dd>key = <b>\"" + tcb->m_key + "\"</b>\n" +
		"<dd>prev index = <a href=\"#"
			+ gm_to_string( tcb->m_prev) + "\">"
			+ gm_to_string( tcb->m_prev) + "</a>\n" +
		"<dd>next index = <a href=\"#"
			+ gm_to_string( tcb->m_next) + "\">"
			+ gm_to_string( tcb->m_next) + "</a>\n" +
		"<dd>m_proc = " + gm_to_string( (void *)m_proc) + "\n";
	return tmp + "</dl>\n</p>\n";
}

ThingControlBlock *ThingManager::get( int id) const
{
	if( (id < 0) || (id >= TM_MAX_THINGS)) return 0;
	// FIXME: Should check if "thing + id" actually exsists. "thing + id"
	//	could be part of the freelist. Perhaps it may be better to
	//	force caller to know if thing exists or not, because this will
	//	simplify the management of things (by not forcing
	//	ThingManager to do it).
	return m_things + id;
}

ThingControlBlock *ThingManager::get( const std::string &key) const
{
	std::map <const std::string, int> :: const_iterator iter =
		m_thing_map.find( key);
	if( iter == m_thing_map.end()) return 0;
	return m_things + (iter->second);
}

void *ThingManager::get_module( const std::string &key) const
{
	std::map <const std::string, void *> :: const_iterator iter =
		m_module.find( key);
	if( iter == m_module.end()) return 0;
	return (iter->second);
}

// http://www.isotton.com/howtos/C++-dlopen-mini-HOWTO/
//	C++-dlopen-mini-HOWTO.html
int ThingManager::load_module( const std::string &key,
	const std::string &fname, std::string &emsg)
{
#ifdef UNIX_LIKE_OS
	if( get_module( key)) return -1;

	emsg.clear();
	// open the shared object library
	void *handle = dlopen( fname.c_str(), RTLD_LAZY);
	if( !handle) {
		emsg += "Cannot open shared object \"" + fname + "\": ";
		const char *p = dlerror();
		if( p) emsg += std::string( p);
		else emsg += "dlerror = no error?";
		return -2;
	}

	// load the symbol
	typedef int (* ModuleMainFn)( const std::string *, GameEngine *,
		std::string *);
	ModuleMainFn gm_main = (ModuleMainFn) dlsym( handle, "gm_main");
	{
		const char *p = dlerror();
		if( p != NULL) {
			emsg += "Cannot load symbol \"gm_main\": " +
				std::string( p);
			dlclose( handle);
			return -3;
		}
	}

	// XXX: UPDATE - Module is now closed on error. Module should cleanup
	//	properly otherwise a SEGFAULT (or worse) could occur.
	// FIXME: If the shared object fails to run its "gm_main" function
	//	then it should be closed (dlclose). However, this requires that
	//	the module perform proper cleanup first before it returned
	//	with an error.
	//
	//	For now, the module will not be closed until the full extent
	//	of the cleanup required is known. Downside is loss in whatever
	//	memory/resources the module took up.
	if( gm_main( &key, m_game_engine, &emsg)) {
		dlclose( handle);
		return -4;
	}

	// Add it to the loaded modules list/map.
	typedef std::pair <const std::string, void *> SPPair;
	if( false == (m_module.insert( SPPair( key, handle)).second)) {
		dlclose( handle);
		return -5;
	}

	return 0;
#else
	emsg = "not implemented for this OS";
	return -1000;
#endif
}
